using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace PrivateDataCommon
{
    /// <summary>
    /// Content - represents any data that we may want to store/retrieve from
    /// the data storage (eventually encrypted into an EncryptedContent).
    /// 
    /// This class is just a container for an open data stream containing the
    /// data in question.
    /// </summary>
    public class Content : IDisposable
    {
        /// <summary>
        /// Constructs an object from a data stream. The content takes
        /// ownership of the stream.
        /// </summary>
        /// <param name="data">Stream containing plain content</param>
        public Content(Stream data)
        {
            this.data = data;
        }

        /// <summary>
        /// Constructs Content from a byte array
        /// </summary>
        /// <param name="dataBytes">Byte array containing plain content</param>
        public Content(byte[] dataBytes)
        {
            MemoryStream stream = new MemoryStream(dataBytes);
            this.data = stream;
        }

        /// <summary>
        /// Destroys the object
        /// </summary>
        ~Content()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// Disposes the resources held by the object (inherited from IDisposable)
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose  resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (data != null)
                {
                    data.Close();
                    data = null;
                }
            }
#if DEBUG
            else
            {
                // If disposing from the destructor, let's throw an exception
                // only if the data hasn't been detached; this means that
                // somebody lost a content that was actually carrying data
                if (data != null)
                {
                    throw new Exception(); // TODO: Fix exceptions
                }
            }
#endif
        }

        /// <summary>
        /// Returns the stream containing the plain content. This detaches the
        /// data from the object, i.e. the object no longer has reference to it.
        /// The caller who detached it is responsible for closing the stream when
        /// done.
        /// </summary>
        /// <returns>Stream containing the plain content, ready to be read</returns>
        public Stream DetachData()
        {
            Stream data = this.data;
            this.data = null;

            // If the stream supports seeking, make sure it's properly positioned
            if (data.CanSeek)
            {
                data.Seek(0, SeekOrigin.Begin);
            }

            return data;
        }

        protected Stream data;
    }
}
