using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace PrivateDataCommon
{
    /// <summary>
    /// Item ID - an ID uniquely identifying each item
    /// </summary>
    public class ItemID : IComparable
    {
        /// <summary>
        /// Constructs ItemID with a new unique ID
        /// </summary>
        public ItemID()
        {
            this.id = Guid.NewGuid();
        }

        /// <summary>
        /// Reads an item ID from a stream
        /// </summary>
        /// <param name="stream">Stream to read from</param>
        public ItemID(Stream stream)
        {
            byte[] guidBytes = new byte[guidLength];
            Utils.FillByteArrayFromStream(guidBytes, stream);
            id = new Guid(guidBytes);
        }

        /// <summary>
        /// Writes the item ID to a stream
        /// </summary>
        /// <param name="stream">Stream to write to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public uint WriteToStream(Stream stream, bool estimate)
        {
            uint bytesWritten = 0;

            byte[] guidBytes = id.ToByteArray();
            if (guidBytes.Length != guidLength)
            {
                // Internal error - why is the GUID not 16 bytes?
                throw new Exception(); // TODO: Fix exceptions
            }

            if (!estimate)
            {
                Utils.WriteByteArrayToStream(guidBytes, stream);
            }
            bytesWritten += (uint)guidBytes.Length;

            return bytesWritten;
        }

        #region IComparable Members

        /// <summary>
        /// Compares the Item ID with another Item ID
        /// </summary>
        /// <param name="obj">Item ID to compare to</param>
        /// <returns>
        /// - Less than 0 if the Item ID is less than the one given
        /// - 0 if the Item ID is equal to the one given
        /// - Greater than 0 if the Item ID is greater than the one given
        /// </returns>
        public int CompareTo(object obj)
        {
            if (obj == null)
            {
                return 1;
            }

            if (obj is ItemID)
            {
                return id.CompareTo((obj as ItemID).ID);
            }

            throw new ArgumentException("Object is not ItemID");
        }

        #endregion

        /// <summary>
        /// Verifies whether the Item ID is equal to another Item ID.
        /// We must override Equals because we're implementing IComparable.
        /// </summary>
        /// <param name="obj">Item ID to compare to</param>
        /// <returns>True if the Item IDs are equal; false otherwise</returns>
        public override bool Equals(object obj)
        {
            if (obj == null)
            {
                return false;
            }
            else
            {
                return (CompareTo(obj) == 0);
            }
        }

        /// <summary>
        /// Returns the object's hash code. We must override this because we're
        /// overriding Equals, and because we might use ItemID objects as keys
        /// in a hash table.
        /// </summary>
        /// <returns>Hash code for the Item ID</returns>
        public override int GetHashCode()
        {
            // Let's just use the hash code of the underlying GUID
            return id.GetHashCode();
        }

        /// <summary>
        /// GUID
        /// </summary>
        public Guid ID
        {
            get
            {
                return id;
            }
        }

        private const int guidLength = 16;
        private Guid id;
    }
}
