using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace PrivateDataCommon
{
    /// <summary>
    /// Various utility methods (mostly stream- and byte array-oriented)
    /// </summary>
    public class Utils
    {
        /// <summary>
        /// Reads an int from a stream
        /// </summary>
        /// <param name="stream">Stream to read the int from</param>
        /// <returns>Int read from the stream</returns>
        public static int ReadIntFromStream(Stream stream)
        {
            byte[] intBytes = new byte[sizeof(int)];

            if (stream.Read(intBytes, 0, intBytes.Length) != intBytes.Length)
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            return BitConverter.ToInt32(intBytes, 0);
        }

        /// <summary>
        /// Writes an int to a stream
        /// </summary>
        /// <param name="i">Int to write</param>
        /// <param name="stream">Stream to write the int to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteIntToStream(int i, Stream stream, bool estimate)
        {
            byte[] intBytes = BitConverter.GetBytes(i);

            if (intBytes.Length != sizeof(int))
            {
                // Internal error: we won't be able to read this back properly
                throw new Exception(); // TODO: Fix exceptions
            }

            if (!estimate)
            {
                stream.Write(intBytes, 0, intBytes.Length);
            }
            return (uint)intBytes.Length;
        }

        /// <summary>
        /// Reads an uint from a stream
        /// </summary>
        /// <param name="stream">Stream to read the uint from</param>
        /// <returns>Uint read from the stream</returns>
        public static uint ReadUIntFromStream(Stream stream)
        {
            byte[] uintBytes = new byte[sizeof(uint)];

            if (stream.Read(uintBytes, 0, uintBytes.Length) != uintBytes.Length)
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            return BitConverter.ToUInt32(uintBytes, 0);
        }

        /// <summary>
        /// Writes an uint to a stream
        /// </summary>
        /// <param name="ui">Uint to write</param>
        /// <param name="stream">Stream to write the uint to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteUIntToStream(uint ui, Stream stream, bool estimate)
        {
            byte[] uintBytes = BitConverter.GetBytes(ui);

            if (uintBytes.Length != sizeof(uint))
            {
                // Internal error: we won't be able to read this back properly
                throw new Exception(); // TODO: Fix exceptions
            }

            if (!estimate)
            {
                stream.Write(uintBytes, 0, uintBytes.Length);
            }
            return (uint)uintBytes.Length;
        }

        /// <summary>
        /// Reads a long from a stream
        /// </summary>
        /// <param name="stream">Stream to read the long from</param>
        /// <returns>Long read from the stream</returns>
        public static long ReadLongFromStream(Stream stream)
        {
            byte[] longBytes = new byte[sizeof(long)];

            if (stream.Read(longBytes, 0, longBytes.Length) != longBytes.Length)
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            return BitConverter.ToInt64(longBytes, 0);
        }

        /// <summary>
        /// Writes a long to a stream
        /// </summary>
        /// <param name="l">Long to write</param>
        /// <param name="stream">Stream to write the long to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteLongToStream(long l, Stream stream, bool estimate)
        {
            byte[] longBytes = BitConverter.GetBytes(l);

            if (longBytes.Length != sizeof(long))
            {
                // Internal error: we won't be able to read this back properly
                throw new Exception(); // TODO: Fix exceptions
            }

            if (!estimate)
            {
                stream.Write(longBytes, 0, longBytes.Length);
            }
            return (uint)longBytes.Length;
        }

        /// <summary>
        /// Reads an ulong from a stream
        /// </summary>
        /// <param name="stream">Stream to read the ulong from</param>
        /// <returns>Ulong read from the stream</returns>
        public static ulong ReadULongFromStream(Stream stream)
        {
            byte[] ulongBytes = new byte[sizeof(ulong)];

            if (stream.Read(ulongBytes, 0, ulongBytes.Length) != ulongBytes.Length)
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            return BitConverter.ToUInt64(ulongBytes, 0);
        }

        /// <summary>
        /// Writes an ulong to a stream
        /// </summary>
        /// <param name="ul">Ulong to write</param>
        /// <param name="stream">Stream to write the ulong to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteULongToStream(ulong ul, Stream stream, bool estimate)
        {
            byte[] ulongBytes = BitConverter.GetBytes(ul);

            if (ulongBytes.Length != sizeof(ulong))
            {
                // Internal error: we won't be able to read this back properly
                throw new Exception(); // TODO: Fix exceptions
            }

            if (!estimate)
            {
                stream.Write(ulongBytes, 0, ulongBytes.Length);
            }
            return (uint)ulongBytes.Length;
        }

        /// <summary>
        /// Reads a bool from a stream
        /// </summary>
        /// <param name="stream">Stream to read the bool from</param>
        /// <returns>Bool read from the stream</returns>
        public static bool ReadBoolFromStream(Stream stream)
        {
            byte[] boolBytes = new byte[sizeof(bool)];

            if (stream.Read(boolBytes, 0, boolBytes.Length) != boolBytes.Length)
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            return BitConverter.ToBoolean(boolBytes, 0);
        }

        /// <summary>
        /// Writes an bool to a stream
        /// </summary>
        /// <param name="b">Bool to write</param>
        /// <param name="stream">Stream to write the bool to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteBoolToStream(bool b, Stream stream, bool estimate)
        {
            byte[] boolBytes = BitConverter.GetBytes(b);

            if (boolBytes.Length != sizeof(bool))
            {
                // Internal error: we won't be able to read this back properly
                throw new Exception(); // TODO: Fix exceptions
            }

            if (!estimate)
            {
                stream.Write(boolBytes, 0, boolBytes.Length);
            }
            return (uint)boolBytes.Length;
        }

        /// <summary>
        /// Reads a string from a stream
        /// </summary>
        /// <param name="stream">Stream to read the string from</param>
        /// <returns>String read from the stream</returns>
        public static string ReadStringFromStream(Stream stream)
        {
            byte[] stringBytes = ReadByteArrayWithLengthFromStream(stream);
            return Encoding.Unicode.GetString(stringBytes, 0, stringBytes.Length);
        }

        /// <summary>
        /// Writes a string to a stream
        /// </summary>
        /// <param name="s">String to write</param>
        /// <param name="stream">Stream to write the string to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteStringToStream(string s, Stream stream, bool estimate)
        {
            byte[] stringBytes = Encoding.Unicode.GetBytes(s);
            return WriteByteArrayWithLengthToStream(stringBytes, stream, estimate);
        }

        /// <summary>
        /// Fills in a pre-allocated byte array from a stream by reading exactly
        /// as many bytes from the stream as the array can hold
        /// </summary>
        /// <param name="bytes">Byte array to fill</param>
        /// <param name="stream">Stream to read the data from</param>
        public static void FillByteArrayFromStream(byte[] bytes, Stream stream)
        {
            if (stream.Read(bytes, 0, bytes.Length) != bytes.Length)
            {
                throw new Exception();
            }
        }

        /// <summary>
        /// Reads a byte array from a stream that contains data prefixed by its
        /// length. ReadByteArrayWithLengthFromStream reads the length, allocates a byte
        /// array of the necessary size, and then fills it with bytes from the
        /// stream.
        /// </summary>
        /// <param name="stream">Stream to read the data from</param>
        /// <returns>Byte array containing the data read</returns>
        public static byte[] ReadByteArrayWithLengthFromStream(Stream stream)
        {
            byte[] bytes = new byte[ReadIntFromStream(stream)];
            FillByteArrayFromStream(bytes, stream);
            return bytes;
        }

        /// <summary>
        /// Writes a byte array to a stream.
        /// </summary>
        /// <param name="bytes">Byte array to write</param>
        /// <param name="stream">Stream to write it to</param>
        public static void WriteByteArrayToStream(byte[] bytes, Stream stream)
        {
            stream.Write(bytes, 0, bytes.Length);
        }

        /// <summary>
        /// Writes a byte array to a stream by writing the array's length
        /// followed by its contents.
        /// </summary>
        /// <param name="bytes">Byte array to write</param>
        /// <param name="stream">Stream to write it to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public static uint WriteByteArrayWithLengthToStream(byte[] bytes, Stream stream, bool estimate)
        {
            uint bytesWritten = 0;

            bytesWritten += WriteIntToStream(bytes.Length, stream, estimate);
            
            if (!estimate)
            {
                WriteByteArrayToStream(bytes, stream);
            }
            bytesWritten += (uint)bytes.Length;

            return bytesWritten;
        }

        /// <summary>
        /// Reads the whole stream
        /// </summary>
        /// <param name="stream">Stream to read</param>
        /// <returns>Byte array containing all the stream's contents</returns>
        public static byte[] ReadStream(Stream stream)
        {
            MemoryStream memoryStream = new MemoryStream();
            try
            {
                byte[] buf = new byte[512];

                int bytesRead = 0;
                do
                {
                    bytesRead = stream.Read(buf, 0, buf.Length);
                    memoryStream.Write(buf, 0, bytesRead);
                }
                while (bytesRead > 0);

                byte[] streamBytes = memoryStream.ToArray();
                return streamBytes;
            }
            finally
            {
                memoryStream.Close();
            }
        }

        /// <summary>
        /// Copies a block of a specified length from a stream to another stream.
        /// </summary>
        /// <param name="source">Source stream</param>
        /// <param name="destination">Destination stream</param>
        /// <param name="length">How many bytes to copy</param>
        /// <param name="controller">Slow operation controller</param>
        public static void CopyFromStreamToStream(
            Stream source,
            Stream destination,
            int length,
            SlowOperationController controller)
        {
            int bytesRemaining = length;
            byte[] buf = new byte[bufferSize];

            while (bytesRemaining > 0)
            {
                controller.ThrowIfCancelRequested();

                int bytesRead = source.Read(buf, 0, Math.Min(buf.Length, bytesRemaining));
                if (bytesRead <= 0)
                {
                    throw new Exception(); // TODO: Fix exceptions
                }

                controller.ThrowIfCancelRequested();

                destination.Write(buf, 0, bytesRead);
                bytesRemaining -= bytesRead;
            }
        }

        /// <summary>
        /// Copies a block of a specified length inside a stream
        /// </summary>
        /// <param name="stream">Stream containing the data</param>
        /// <param name="sourceOffset">Source offset</param>
        /// <param name="destinationOffset">
        /// Destination offset. Should be outside the source range.
        /// </param>
        /// <param name="length">How many bytes to copy</param>
        /// <param name="controller">Slow operation controller</param>
        public static void CopyInsideStream(
            Stream stream,
            ulong sourceOffset,
            ulong destinationOffset,
            uint length,
            SlowOperationController controller)
        {
            if (destinationOffset > sourceOffset
                    && destinationOffset < sourceOffset + length)
            {
                throw new ArgumentException(); // TODO: Fix exceptions
            }

            ulong currentSourceOffset = sourceOffset;
            ulong currentDestinationOffset = destinationOffset;
            uint bytesRemaining = length;
            byte[] buf = new byte[bufferSize];

            while (bytesRemaining > 0)
            {
                controller.ThrowIfCancelRequested();

                stream.Seek((long)currentSourceOffset, SeekOrigin.Begin);
                uint bytesRead
                    = (uint)stream.Read(
                        buf,
                        0,
                        (int)Math.Min((uint)buf.Length, bytesRemaining));
                if (bytesRead <= 0)
                {
                    throw new Exception(); // TODO: Fix exceptions
                }
                currentSourceOffset += bytesRead;

                controller.ThrowIfCancelRequested();

                stream.Seek((long)currentDestinationOffset, SeekOrigin.Begin);
                stream.Write(buf, 0, (int)bytesRead);
                currentDestinationOffset += bytesRead;

                bytesRemaining -= bytesRead;
            }
        }

        /// <summary>
        /// Copies all data from the current position in a source stream to its
        /// length into a destination stream.
        /// </summary>
        /// <param name="source">Source stream to copy data from</param>
        /// <param name="destination">Destination stream to copy data to</param>
        /// <returns>Number of bytes copied</returns>
        public static int CopyWholeStreamToStream(
            Stream source,
            Stream destination,
            SlowOperationController controller)
        {
            byte[] buf = new byte[bufferSize];

            int totalBytesCopied = 0;
            int bytesRead = 0;
            do
            {
                controller.ThrowIfCancelRequested();

                bytesRead = source.Read(buf, 0, buf.Length);
                if (bytesRead > 0)
                {
                    controller.ThrowIfCancelRequested();

                    destination.Write(buf, 0, bytesRead);
                    totalBytesCopied += bytesRead;
                }
            }
            while (bytesRead > 0);

            return totalBytesCopied;
        }


        /// <summary>
        /// Compares the contents of two byte arrays
        /// </summary>
        /// <param name="array1">First array to compare</param>
        /// <param name="array2">Second array to compare</param>
        /// <returns>True if the arrays' contents match. Otherwise false.</returns>
        public static bool ByteArraysMatch(byte[] array1, byte[] array2)
        {
            if (array1.Length != array2.Length)
            {
                return false;
            }

            for (int i = 0; i < array1.Length; i++)
            {
                if (array1[i] != array2[i])
                {
                    return false;
                }
            }

            return true;
        }

        private static int bufferSize = 512;
    }
}
