using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataCore
{
    internal enum ArchiveFileMode
    {
        Archive,
        Restore
    }

    internal class ArchiveFileManager
    {
        private enum StoredItemType
        {
            CustomData = 0,
            Item = 1
        }

        public ArchiveFileManager(string fileName, ArchiveFileMode backupMode)
        {
            this.backupMode = backupMode;
            
            // Create the file
            switch (backupMode)
            {
                case ArchiveFileMode.Archive:
                    fileStream = new FileStream(
                        fileName,
                        FileMode.CreateNew,
                        FileAccess.Write,
                        FileShare.None);
                    break;

                case ArchiveFileMode.Restore:
                    fileStream = new FileStream(
                        fileName,
                        FileMode.Open,
                        FileAccess.Read,
                        FileShare.Read);
                    break;

                default:
                    // Internal error - unknown file mode
                    throw new Exception(); // TODO: Fix exceptions
            }
        }

        public void ReadBackupFile(SlowOperationController controller)
        {
            controller.StartProgress((ulong)fileStream.Length); // TODO: remove cast
            while (fileStream.Position < fileStream.Length)
            {
                ReadNext(controller);
                controller.ReportProgress((ulong)fileStream.Position);
            }
            controller.FinishProgress();
        }

        private void ReadNext(SlowOperationController controller)
        {
            byte storedItemType = (byte)fileStream.ReadByte();

            switch ((StoredItemType)storedItemType)
            {
                case StoredItemType.CustomData:
                    ReadCustomData(controller);
                    break;

                case StoredItemType.Item:
                    ReadItem(controller);
                    break;

                default:
                    // File is corrupted
                    throw new Exception(); // TODO: Fix exceptions
            }
        }

        private void WriteContent(Content content, SlowOperationController controller)
        {
            // If encrypted - store the IV
            if (content is EncryptedContent)
            {
                Utils.WriteByteArrayWithLengthToStream(
                    ((EncryptedContent)content).IV.Bytes,
                    fileStream,
                    false/*estimate*/);
            }

            Stream contentStream = content.DetachData();
            try
            {
                // Leave space for the content's length
                long lengthPosition = fileStream.Position;
                Utils.WriteIntToStream(0, fileStream, false/*estimate*/);

                // Write the content
                int length
                    = Utils.CopyWholeStreamToStream(
                        contentStream,
                        fileStream,
                        controller);

                // Go back and write the length
                fileStream.Seek(lengthPosition, SeekOrigin.Begin);
                Utils.WriteIntToStream(length, fileStream, false/*estimate*/);

                // And go back to where we were
                fileStream.Seek(length, SeekOrigin.Current);
            }
            finally
            {
                contentStream.Close();
            }
        }

        private Content ReadContent(bool encrypted)
        {
            // If encrypted - read the IV
            Content content;
            byte[] contentBytes;

            // Read IV if needed
            byte[] ivBytes = null;
            if (encrypted)
            {
                ivBytes = Utils.ReadByteArrayWithLengthFromStream(fileStream);
            }

            // TODO: Large items? We can't just read them in memory...
            contentBytes
                = Utils.ReadByteArrayWithLengthFromStream(fileStream);

            // Create Content object
            if (encrypted)
            {
                content = new EncryptedContent(
                    new MemoryStream(contentBytes),
                    new IV(ivBytes));
            }
            else
            {
                content = new Content(contentBytes);
            }

            return content;
        }

        public void WriteCustomData(
            int type,
            Content content,
            SlowOperationController controller)
        {
            // We're storing custom data
            fileStream.WriteByte((byte)StoredItemType.CustomData);

            // Type
            Utils.WriteIntToStream(type, fileStream, false/*estimate*/);

            // Is the content encrypted?
            Utils.WriteBoolToStream(
                content is EncryptedContent,
                fileStream,
                false/*estimate*/);

            // Content
            WriteContent(content, controller);
        }

        private void ReadCustomData(SlowOperationController controller)
        {
            // Type
            int type = Utils.ReadIntFromStream(fileStream);

            // Is the content encrypted?
            bool encrypted = Utils.ReadBoolFromStream(fileStream);

            // Content
            Content content = ReadContent(encrypted);

            // Call the data manager to restore the data we just read
            OnReadCustomData(
                new CustomDataReadEventArgs(type, content, controller));

            content.Dispose();
        }

        public void WriteItem(
            ItemID id,
            EncryptedContent encryptedMetadata,
            EncryptedContent encryptedContent,
            SlowOperationController controller)
        {
            // We're storing a private data item
            fileStream.WriteByte((byte)StoredItemType.Item);

            controller.StartProgress(3);
            try
            {
                // ID
                id.WriteToStream(fileStream, false/*estimate*/);

                controller.ReportProgress(1);

                // Metadata
                WriteContent(encryptedMetadata, controller);

                controller.ReportProgress(2);

                // Content
                WriteContent(encryptedContent, controller);
            }
            finally
            {
                controller.FinishProgress();
            }
        }

        private void ReadItem(SlowOperationController controller)
        {
            // ID
            ItemID id = new ItemID(fileStream);

            // Metadata
            EncryptedContent encryptedMetadata
                = (EncryptedContent)ReadContent(true/*encrypted*/);

            // Content
            EncryptedContent encryptedContent
                = (EncryptedContent)ReadContent(true/*encrypted*/);

            OnReadItem(
                new ItemReadEventArgs(
                    id,
                    encryptedMetadata,
                    encryptedContent,
                    controller));

            encryptedMetadata.Dispose();
            encryptedContent.Dispose();
        }

        public class ItemReadEventArgs : EventArgs
        {
            public ItemReadEventArgs(
                ItemID id,
                EncryptedContent metadata,
                EncryptedContent content,
                SlowOperationController controller)
            {
                this.id = id;
                this.metadata = metadata;
                this.content = content;

                this.controller = controller;
            }

            public ItemID ID
            {
                get
                {
                    return id;
                }
            }

            public EncryptedContent Metadata
            {
                get
                {
                    return metadata;
                }
            }

            public EncryptedContent Content
            {
                get
                {
                    return content;
                }
            }

            public SlowOperationController Controller
            {
                get
                {
                    return controller;
                }
            }

            private ItemID id;
            private EncryptedContent metadata;
            private EncryptedContent content;
            
            private SlowOperationController controller;
        }

        public class CustomDataReadEventArgs : EventArgs
        {
            public CustomDataReadEventArgs(
                int customDataType,
                Content content,
                SlowOperationController controller)
            {
                this.customDataType = customDataType;
                this.content = content;

                this.controller = controller;
            }

            public int CustomDataType
            {
                get
                {
                    return customDataType;
                }
            }

            public Content Content
            {
                get
                {
                    return content;
                }
            }

            public SlowOperationController Controller
            {
                get
                {
                    return controller;
                }
            }

            private int customDataType;
            private Content content;

            private SlowOperationController controller;
        }

        protected void OnReadItem(ItemReadEventArgs e)
        {
            if (ItemRead != null)
            {
                ItemRead(this, e);
            }
        }

        protected void OnReadCustomData(CustomDataReadEventArgs e)
        {
            if (CustomDataRead != null)
            {
                CustomDataRead(this, e);
            }
        }

        public event EventHandler<ItemReadEventArgs> ItemRead;
        public event EventHandler<CustomDataReadEventArgs> CustomDataRead;

        private FileStream fileStream;
        private ArchiveFileMode backupMode;
    }
}
