using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Security.Cryptography;

using PrivateDataCommon;

namespace PrivateDataCore
{
    /// <summary>
    /// Wrapper around the specific cryptographic implementation being used
    /// </summary>
    internal class Crypto
    {
        /// <summary>
        /// This class has only static methods
        /// </summary>
        private Crypto()
        {
        }

        /// <summary>
        /// Generates a random key
        /// </summary>
        /// <returns>Random key</returns>
        public static DataKey GenerateKey()
        {
            SymmetricAlgorithm cipher = SymmetricAlgorithm.Create(encryptionAlgorithm);
            return new DataKey(cipher.Key);
        }

        /// <summary>
        /// Generates a random salt suitable for use with the password-based
        /// key derivation functions (KeyFromPassword/PBKDF2).
        /// </summary>
        /// <returns>Randomly generated salt</returns>
        public static byte[] GenerateSalt()
        {
            // Anything random will do, as long as it is bigger than 64 bits
            // (see PKCS #5 v2.0). We'll just use the cipher to generate one
            // for us.
            SymmetricAlgorithm cipher = SymmetricAlgorithm.Create(encryptionAlgorithm);
            return cipher.Key;
        }

        /// <summary>
        /// Derives a cryptographic key from a given password
        /// </summary>
        /// <param name="password">Password to derive key from</param>
        /// <param name="salt">Salt to be used for randomizing the key</param>
        /// <param name="iterationCount">How many iterations of the derivation algorithm?</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>Key suitable for encryption purposes</returns>
        public static DataKey KeyFromPassword(
            string password,
            byte[] salt,
            uint iterationCount,
            SlowOperationController controller)
        {
            byte[] key
                = PBKDF2(
                    password,
                    salt,
                    iterationCount,
                    (uint)SymmetricAlgorithm.Create(encryptionAlgorithm).KeySize / 8,
                    controller);

            return new DataKey(key);
        }

        /// <summary>
        /// Derives a verifier from the key derived from the password. This
        /// verifier can be compared to the one saved in the storage in order
        /// to verify whether the password supplied by the user is correct.
        /// 
        /// We need to use the password's key instead of the password itself
        /// so that generating a verifier from the password is at least as slow
        /// as generating the master data key from the password. We need this
        /// so that we don't make it easier for attackers to brute force the
        /// password against the verifier (it would be easier to brute force it
        /// against the key itself).
        /// </summary>
        /// <param name="passwordKey">Key derived from the password by PBKDF2</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>Password verifier</returns>
        public static byte[] VerifierFromPassword(
            DataKey passwordKey,
            SlowOperationController controller)
        {
            return PBKDF2(
                        passwordKey.Bytes,
                        passwordVerifierSalt,
                        1,
                        passwordVerifierLength,
                        controller);
        }

        /// <summary>
        /// Verifies that a password (given here by the PBKDF2 key derived from
        /// it) is the correct one used to derive the verifier.
        /// </summary>
        /// <param name="passwordKey">
        /// Key derived from the password by PBKDF2. See VerifierFromPassword.
        /// </param>
        /// <param name="verifier">Verifier to validate the password against</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>True if the password is correct; false if not</returns>
        public static bool VerifyPassword(
            DataKey passwordKey,
            byte[] verifier,
            SlowOperationController controller)
        {
            return Utils.ByteArraysMatch(
                    VerifierFromPassword(passwordKey, controller),
                    verifier);
        }

        /// <summary>
        /// Creates EncryptedContent containing PlainContent's data encrypted
        /// with the specified key. Notice that no encryption actually happens
        /// here - this just wraps an encrypting stream around the plain
        /// content's stream. Encryption will happen on-demand as data is being
        /// read from the EncryptedContent object.
        /// </summary>
        /// <param name="content">Plain content we want encrypted</param>
        /// <param name="key">Key to encrypt with</param>
        /// <returns>Encrypted content</returns>
        public static EncryptedContent Encrypt(Content content, DataKey key)
        {
            SymmetricAlgorithm cipher = SymmetricAlgorithm.Create(encryptionAlgorithm);
            IV iv = new IV(cipher.IV);
            Stream cryptoStream 
                = new CryptoStream(content.DetachData(),
                                   cipher.CreateEncryptor(key.Bytes, iv.Bytes),
                                   CryptoStreamMode.Read);

            return new EncryptedContent(cryptoStream, iv);
        }

        /// <summary>
        /// Creates PlainContent decrypting an EncryptedContent with the
        /// specified key. Notice that no decryption actually hapens here -
        /// it just wraps a decrypting stream around the encrypted content's
        /// stream. Decryption will happen on-demand as data is being read from
        /// the PlainContent object.
        /// </summary>
        /// <param name="content">Encrypted content we want decrypted</param>
        /// <param name="key">Key to decrypt with</param>
        /// <returns>Plain content</returns>
        public static Content Decrypt(EncryptedContent content, DataKey key)
        {
            SymmetricAlgorithm cipher = SymmetricAlgorithm.Create(encryptionAlgorithm);
            Stream cryptoStream
                = new CryptoStream(content.DetachData(),
                                   cipher.CreateDecryptor(key.Bytes, content.IV.Bytes),
                                   CryptoStreamMode.Read);

            return new Content(cryptoStream);
        }

        /// <summary>
        /// Keyed-Hashing for Message Authentication (per RFC 2104).
        /// 
        /// Uses a cryptographic hash function in combination with a secret
        /// key for the purposes of message authentication. However, this can
        /// also be used as a pseudorandom function, as described in section B.1.1
        /// of the PKCS #5 2.0 specification.
        /// </summary>
        /// <param name="hash">Hash to be used</param>
        /// <param name="key">Secret shared key to use for hashing the text</param>
        /// <param name="text">Text to be hashed</param>
        /// <returns>Keyed hash of the text</returns>
        private static byte[] HMAC(HashAlgorithm hash, byte[] key, byte[] text)
        {
            const byte innerPadding = 0x36;
            const byte outerPadding = 0x5c;

            byte[] normalizedKey = NormalizeKey(key, hash);

            byte[] innerHash = HashKeyedData(hash, normalizedKey, innerPadding, text);
            byte[] outerHash = HashKeyedData(hash, normalizedKey, outerPadding, innerHash);

            return outerHash;
        }

        /// <summary>
        /// Normalizes the length of a key to match the input block size of a
        /// hash, per RFC 2104:
        ///     1. If the key is shorter than the block size, just pad with zeros
        ///     2. If the key is longer, hash it first; then proceed with step 1
        /// </summary>
        /// <param name="key">Key to be normalized</param>
        /// <param name="hash">Hash algorithm to normalize the key for</param>
        /// <returns>Key whose length matches the hash algorithm's input block size</returns>
        private static byte[] NormalizeKey(byte[] key, HashAlgorithm hash)
        {
            byte[] normalizedKey = new byte[512/8]; // HACK
            if (key.Length <= normalizedKey.Length)
            {
                Array.Copy(key, normalizedKey, key.Length);
            }
            else
            {
                byte[] keyHash = hash.ComputeHash(key);
                Array.Copy(keyHash, normalizedKey, keyHash.Length);
            }

            return normalizedKey;
        }

        /// <summary>
        /// Computes the hash of data appended to a given key (the key being
        /// XOR-ed with a given padding), per RFC 2104.
        /// </summary>
        /// <param name="hash">Hash algorithm to use</param>
        /// <param name="key">Key to prepend to the data</param>
        /// <param name="padding">Padding byte, used to XOR the key with</param>
        /// <param name="data">Data whose keyed hash is to be calculated</param>
        /// <returns>Hash of the data combined with the key</returns>
        private static byte[] HashKeyedData(HashAlgorithm hash, byte[] key, byte padding, byte[] data)
        {
            byte[] paddedData = new byte[key.Length + data.Length];
            for (int i = 0; i < key.Length; i++)
            {
                paddedData[i] = (byte)(key[i] ^ padding);
            }
            Array.Copy(data, 0, paddedData, key.Length, data.Length);

            return hash.ComputeHash(paddedData);
        }

        /// <summary>
        /// Implementation of the Password-Based Key Derivation Function 2,
        /// as specified in PKCS #5 v2.0
        /// </summary>
        /// <param name="password">Password to derive a key from</param>
        /// <param name="salt">Salt used to randomize the key</param>
        /// <param name="iterationCount">How many iterations of the hash
        /// to apply? More means stronger key but slow generation.</param>
        /// <param name="keyLength">Length in bytes of the desired key</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>Key derived from the password</returns>
        private static byte[] PBKDF2(
            string password,
            byte[] salt,
            uint iterationCount,
            uint keyLength,
            SlowOperationController controller)
        {
            byte[] passwordBytes = Encoding.UTF8.GetBytes(password);
            return PBKDF2(passwordBytes, salt, iterationCount, keyLength, controller);
        }

        /// <summary>
        /// Implementation of the Password-Based Key Derivation Function 2,
        /// as specified in PKCS #5 v2.0. This is the core implementation,
        /// when the password has already been converted to a byte array.
        /// </summary>
        /// <param name="passwordBytes">Password, already converted to byte array</param>
        /// <param name="salt"></param>
        /// <param name="iterationCount">How many iterations of the hash
        /// to apply? More means stronger key but slow generation.</param>
        /// <param name="keyLength">Length in bytes of the desired key</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>Key derived from the password</returns>
        private static byte[] PBKDF2(
            byte[] passwordBytes,
            byte[] salt,
            uint iterationCount,
            uint keyLength,
            SlowOperationController controller)
        {
            // Hash to use with the HMAC algorithm as a pseudo-random function
            HashAlgorithm hash = HashAlgorithm.Create(hashAlgorithm);
            uint hashLength = (uint)hash.HashSize / 8;

            // (keyLength / hashLength), rounded up
            uint octetBlocks = (keyLength + hashLength - 1) / hashLength;
            uint bytesInLastBlock = keyLength % hashLength;
            byte[] outputBytes = new byte[octetBlocks * hashLength];

            controller.StartProgress(octetBlocks);
            try
            {
                for (uint block = 0; block < octetBlocks; block++)
                {
                    controller.ThrowIfCancelRequested();
                    controller.ReportProgress(block);

                    // Combine the salt with the current block's index so that each
                    // block can have its own salt
                    byte[] blockSalt = new byte[salt.Length + sizeof(int)];
                    {
                        // The block's salt has the common salt at its beginning
                        Array.Copy(salt, blockSalt, salt.Length);

                        // Followed by a big-endian representation of the block's index
                        for (int b = 0; b < sizeof(int); b++)
                        {
                            blockSalt[salt.Length + b]
                                = (byte)(block >> ((sizeof(int) - b - 1) * 8));
                        }
                    }

                    // TODO: From resource
                    controller.StartProgress(iterationCount);
                    try
                    {
                        // Perform the specified number of iterations of the pseudo-
                        // random function (in our case - the HMAC algorithm) on the
                        // password combined with the salt
                        byte[] blockHash = HMAC(hash, passwordBytes, blockSalt);
                        for (uint iteration = 1; iteration < iterationCount; iteration++)
                        {
                            // These can be time consuming, so let's not do
                            // them on every iteration
                            if (iteration % 256 == 0)
                            {
                                controller.ThrowIfCancelRequested();
                                controller.ReportProgress(iteration);
                            }

                            byte[] tempHash = HMAC(hash, passwordBytes, blockHash);
                            for (uint i = 0; i < blockHash.Length; i++)
                            {
                                blockHash[i] ^= tempHash[i];
                            }
                        }

                        Array.Copy(
                            blockHash,
                            0,
                            outputBytes,
                            (int)(block * hashLength),
                            (int)hashLength);
                    }
                    finally
                    {
                        controller.FinishProgress();
                    }
                }
            }
            finally
            {
                controller.FinishProgress();
            }

            // Cut the calculated hash data to the desired length and return it
            byte[] outputKey = new byte[keyLength];
            Array.Copy(outputBytes, outputKey, outputKey.Length);
            return outputKey;
        }

        // Encryption and hash algorithms that Private Data will use
        private const string encryptionAlgorithm = "Rijndael";
        private const string hashAlgorithm = "SHA1";

        private static byte[] passwordVerifierSalt = {1, 2, 3}; // TODO: make it really random!
        private const int passwordVerifierLength = 32;
    }
}
