using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;
using PrivateDataStorage;

namespace PrivateDataCore
{
    /// <summary>
    /// Data Manager - manages the data flow
    /// </summary>
    public class DataManager
    {
        /// <summary>
        /// Creates a brand new storage under a specified data path.
        /// </summary>
        /// <param name="dataPath">Path to the store</param>
        /// <param name="password">Initial password</param>
        /// <param name="iterationCount">
        /// Iteration count to use with the password derivation function
        /// </param>
        /// <param name="controller">Slow operation controller</param>
        public void CreateStorage(
            string dataPath,
            string password,
            uint iterationCount,
            SlowOperationController controller)
        {
            controller.StartProgress(2);

            // Create an empty storage and close it so that we can log into it
            DataStorage storage = DataStorage.CreateStorage(dataPath);
            try
            {
                EncryptionHeader encryptionHeader
                    = new EncryptionHeader(
                        password,
                        iterationCount,
                        controller);
                storage.StoreCustomData(
                    (int)CustomDataType.EncryptionHeader,
                    encryptionHeader.ToContent());
            }
            finally
            {
                storage.Close();
            }

            controller.ReportProgress(1);

            // Log into the storage we just created
            if (!Login(dataPath, password, controller))
            {
                // Internal error: we couldn't log into the just created storage!!!
                throw new Exception(); // TODO: Fix exceptions
            }

            controller.FinishProgress();
        }

        /// <summary>
        /// Logs into a specified data storage
        /// </summary>
        /// <param name="dataPath">Path of the store</param>
        /// <param name="password">Password for the particular storage</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>
        /// True if login is successful. False if the password is incorrect
        /// (will throw exception if the data location is not valid).
        /// </returns>
        public bool Login(
            string dataPath,
            string password,
            SlowOperationController controller)
        {
            storage = new DataStorage(dataPath);

            Content headerContent
                = storage.GetCustomData((int)CustomDataType.EncryptionHeader);
            EncryptionHeader header = new EncryptionHeader(headerContent);
            headerContent.Dispose();

            try
            {
                // Obtain the data key
                try
                {
                    key = header.GetDataKey(password, controller);
                }
                catch // TODO: Catch wrong password exception
                {
                    // We need to close the storage we just opened on failure!
                    storage.Close();
                    return false;
                }

                // Now the storage is properly open (we have the key);
                // let's read in the folder structure
                itemTree = new ItemTree(this, storage.GetItemIDs());
            }
            catch
            {
                // We need to close the storage we just opened on failure!
                storage.Close();
                throw;
            }

            return true;
        }

        public void Close()
        {
            storage.Close();
        }

        /// <summary>
        /// Retrieves the metadata for a particular item
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>Metadata for the item</returns>
        public Metadata GetMetadata(ItemID id)
        {
            EncryptedContent encryptedMetadata = storage.GetMetadata(id);
            try
            {
                Content decryptedMetadata = Crypto.Decrypt(encryptedMetadata, key);
                try
                {
                    return new Metadata(decryptedMetadata);
                }
                finally
                {
                    decryptedMetadata.Dispose();
                }
            }
            finally
            {
                encryptedMetadata.Dispose();
            }
        }

        /// <summary>
        /// Retrieves the contents of a particular private item
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>Item's contents</returns>
        public Item GetItem(ItemID id)
        {
            Content decryptedContent;

            EncryptedContent encryptedContent = storage.GetContent(id);
            try
            {
                decryptedContent = Crypto.Decrypt(encryptedContent, key);
            }
            finally
            {
                encryptedContent.Dispose();
            }

            try
            {
                return Item.Create(id, GetMetadata(id), decryptedContent);
            }
            finally
            {
                decryptedContent.Dispose();
            }
        }

        /// <summary>
        /// Stores the metadata for an existing private item.
        /// This metod is private because there is no need to export it.
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <param name="metadata">Metadata to store for the item</param>
        /// <param name="controller">Slow operation controller</param>
        private void StoreMetadata(
            ItemID id,
            Metadata metadata,
            SlowOperationController controller)
        {
            EncryptedContent encryptedMetadata = Crypto.Encrypt(metadata.ToContent(), key);
            storage.StoreMetadata(id, encryptedMetadata, controller);
        }

        /// <summary>
        /// Saves a private item to the storage
        /// </summary>
        /// <param name=param name="item">Item to be stored</param>
        /// <param name="folder">Folder to put the item in (null for root)</param>
        /// <param name="controller">Slow operation status</param>
        public void StoreItem(
            Item item,
            ItemID folder,
            SlowOperationController controller)
        {
            EncryptedContent encryptedMetadata = null;
            EncryptedContent encryptedContent = null;
            try
            {
                controller.StartProgress(2);

                // Encrypt the item's metadata and contents
                {
                    Content metadata = item.GenerateMetadata();
                    Content content = item.GenerateContent();
                    try
                    {
                        encryptedMetadata = Crypto.Encrypt(metadata, key);
                        encryptedContent = Crypto.Encrypt(content, key);
                    }
                    finally
                    {
                        metadata.Dispose();
                        content.Dispose();
                    }
                }

                controller.ThrowIfCancelRequested();

                // Send the encrypted data to the storage
                storage.StoreItem(
                    item.ID,
                    encryptedMetadata,
                    encryptedContent,
                    item is IFileItem,
                    controller);
                OnItemStored(new ItemEventArgs(item.ID));

                controller.ReportProgress(1);

                // And file the item in the appropriate folder
                bool rootChanged;
                itemTree.MoveItemToFolder(
                    item.ID,
                    folder,
                    out rootChanged,
                    controller);
                if (rootChanged)
                {
                    // We have several cases:
                    // 1. We created a new item and moved it into the root
                    //    (despite the source and destination folders being the
                    //    same, the flag does get set because the item did not
                    //    exist)
                    //      => we need to update the UI for the root
                    // 2. We stored an item that already existed in the root
                    //    and moved it away (if we just created a new item, the
                    //    rootChanged flag does not get set)
                    //      => we need to update the UI for the root
                    // 3. We stored an item that already existed in another
                    //    folder and we moved it into the root
                    //      => we need to update the UI for the root
                    //
                    // Bottom line: we need to update the UI for the root in
                    // all cases that the flag gets set
                    OnItemStored(new ItemEventArgs(null));
                }

                controller.FinishProgress();
            }
            finally
            {
                if (encryptedMetadata != null)
                {
                    encryptedMetadata.Dispose();
                }

                if (encryptedContent != null)
                {
                    encryptedContent.Dispose();
                }
            }
        }

        /// <summary>
        /// Deletes a private item
        /// </summary>
        /// <param name="id">ID of the item to delete</param>
        /// <param name="controller">Slow operation controller</param>
        public void DeleteItem(ItemID id, SlowOperationController controller)
        {
            storage.DeleteItem(id);
            itemTree.DeleteItem(id, controller);
            OnItemDeleted(new ItemEventArgs(id));
        }

        /// <summary>
        /// Renames an item
        /// </summary>
        /// <param name="id">ID of the item to rename</param>
        /// <param name="name">New name</param>
        /// <param name="controller">Slow operation controller</param>
        public void RenameItem(
            ItemID id,
            string name,
            SlowOperationController controller)
        {
            Metadata metadata = GetMetadata(id);
            metadata.Name = name;
            StoreMetadata(id, metadata, controller);

            OnItemStored(new ItemEventArgs(id));
        }

        public ItemID[] GetFolderItems(ItemID folder)
        {
            return itemTree.GetFolderItems(folder);
        }

        /// <summary>
        /// Moves an item into a specified folder (or puts a new item into a
        /// specified folder)
        /// </summary>
        /// <param name="item">Item to file into the folder</param>
        /// <param name="folder">Folder to file the item into</param>
        /// <param name="controller">Slow operation controller</param>
        public void MoveItemToFolder(
            ItemID item,
            ItemID folder,
            SlowOperationController controller)
        {
            bool rootChanged;
            itemTree.MoveItemToFolder(item, folder, out rootChanged, controller);
            if (rootChanged)
            {
                OnItemStored(new ItemEventArgs(null));
            }
        }

        public void ChangePassword(
            string oldPassword,
            string newPassword,
            uint newIterationCount,
            SlowOperationController controller)
        {
            // Obtain the encryption header
            EncryptionHeader encryptionHeader
                = new EncryptionHeader(
                    storage.GetCustomData((int)CustomDataType.EncryptionHeader));

            // Re-encrypt the data key with the new password
            encryptionHeader.ChangePassword(
                oldPassword,
                newPassword,
                newIterationCount, controller);

            // And store the header back
            storage.StoreCustomData(
                (int)CustomDataType.EncryptionHeader,
                encryptionHeader.ToContent());
        }

        public void BackupData(
            string fileName,
            string password,
            uint iterationCount,
            SlowOperationController controller)
        {
            // TODO: Implement
            ArchiveFileManager backup = new ArchiveFileManager(fileName, ArchiveFileMode.Archive);
            backupKey = Crypto.GenerateKey();

            ItemID[] items = storage.GetItemIDs();
            controller.StartProgress((uint)items.Length + 3);

            // Create new encryption header
            EncryptionHeader encryptionHeader
                = new EncryptionHeader(password, iterationCount, controller);
            controller.ReportProgress(1);
            backupKey = encryptionHeader.GetDataKey(password, controller);
            controller.ReportProgress(2);


            // Store the encryption header
            backup.WriteCustomData(
                (int)CustomDataType.EncryptionHeader,
                encryptionHeader.ToContent(),
                controller);
            controller.ReportProgress(3);

            uint count = 3;
            foreach (ItemID id in items)
            {
                backup.WriteItem(
                    id,
                    Crypto.Encrypt(
                        Crypto.Decrypt(storage.GetMetadata(id), key),
                        backupKey),
                    Crypto.Encrypt(
                        Crypto.Decrypt(storage.GetContent(id), key),
                        backupKey),
                    controller);
                controller.ReportProgress(++count);
            }
            controller.FinishProgress();
        }

        public void RestoreData(
            string fileName,
            string password,
            SlowOperationController controller)
        {
            ArchiveFileManager backup = new ArchiveFileManager(fileName, ArchiveFileMode.Restore);
            this.backupPassword = password;

            backup.ItemRead
                += new EventHandler<ArchiveFileManager.ItemReadEventArgs>(
                    backup_ReadItem);
            backup.CustomDataRead
                += new EventHandler<ArchiveFileManager.CustomDataReadEventArgs>(
                    backup_ReadCustomData);

            backup.ReadBackupFile(controller);
            itemTree = null;
            itemTree = new ItemTree(this, storage.GetItemIDs());
            OnItemTreeReloaded(new EventArgs());
        }

        private void backup_ReadItem(
            object sender,
            ArchiveFileManager.ItemReadEventArgs e)
        {
            if (backupKey != null)
            {
                Metadata metadata
                    = new Metadata(Crypto.Decrypt(e.Metadata, backupKey));
                storage.StoreItem(
                    e.ID,
                    Crypto.Encrypt(metadata.ToContent(), key),
                    Crypto.Encrypt(Crypto.Decrypt(e.Content, backupKey), key),
                    metadata.Type == ItemType.File, 
                    e.Controller);
                OnItemStored(new ItemEventArgs(e.ID));
            }
        }

        private void backup_ReadCustomData(
            object sender,
            ArchiveFileManager.CustomDataReadEventArgs e)
        {
            if ((CustomDataType)e.CustomDataType == CustomDataType.EncryptionHeader)
            {
                EncryptionHeader encryptionHeader
                    = new EncryptionHeader(e.Content);
                backupKey
                    = encryptionHeader.GetDataKey(backupPassword, e.Controller);
                backupPassword = string.Empty;
            }
        }

        public void Defragment(SlowOperationController controller)
        {
            storage.Defragment(controller);
        }

        protected void OnItemTreeReloaded(EventArgs e)
        {
            if (ItemTreeReloaded != null)
            {
                ItemTreeReloaded(this, e);
            }
        }

        protected void OnItemStored(ItemEventArgs e)
        {
            if (ItemStored != null)
            {
                ItemStored(this, e);
            }
        }

        protected void OnItemDeleted(ItemEventArgs e)
        {
            if (ItemDeleted != null)
            {
                ItemDeleted(this, e);
            }
        }

        public class ItemEventArgs : EventArgs
        {
            public ItemEventArgs(ItemID item)
            {
                this.item = item;
            }

            public ItemID Item
            {
                get
                {
                    return item;
                }
            }

            ItemID item;
        }

        public static void TestLoginSpeed(
            uint iterationCount,
            SlowOperationController controller)
        {
            string password = "uhiu5y8t84t02t8y)*&&h^&%bg^%#p9vb79VO87N";
            byte[] salt = Crypto.GenerateSalt();
            Crypto.KeyFromPassword(password, salt, iterationCount, controller);
        }

        public event EventHandler<ItemEventArgs> ItemStored;
        public event EventHandler<ItemEventArgs> ItemDeleted;

        public event EventHandler<EventArgs> ItemTreeReloaded;

        /// <summary>
        /// Enumeration of the possible types of custom data the storage supports
        /// </summary>
        private enum CustomDataType
        {
            EncryptionHeader
        }

        // Master key for access to the encrypted private data
        private DataKey key;

        // The key used for backup
        private DataKey backupKey;

        // Storage for the private items
        private DataStorage storage;

        // Representation of the tree structure of the items
        private ItemTree itemTree;

        private string backupPassword;
    }
}
