using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;
using PrivateDataCore.Exceptions;

namespace PrivateDataCore
{
    /// <summary>
    /// Encryption Header. Stored to and retrieved from the storage, it contains
    /// secure information necessary to set up encryption and decryption
    /// of the private data.
    /// </summary>
    internal class EncryptionHeader
    {
        /// <summary>
        /// Constructs EncryptionHeader from a its contents
        /// </summary>
        /// <param name="header">Contents of the header</param>
        public EncryptionHeader(Content headerContent)
        {
            Stream headerStream = headerContent.DetachData();

            try
            {
                passwordSalt = Utils.ReadByteArrayWithLengthFromStream(headerStream);
                iterationCount = Utils.ReadUIntFromStream(headerStream);
                passwordVerifier = Utils.ReadByteArrayWithLengthFromStream(headerStream);
                encryptedDataKey = Utils.ReadByteArrayWithLengthFromStream(headerStream);
                dataKeyIV = Utils.ReadByteArrayWithLengthFromStream(headerStream);
            }
            finally
            {
                headerStream.Close();
            }
        }

        /// <summary>
        /// Constructs a new EncryptionHeader suitable for use when creating a
        /// new data storage. All elements that should be random (keys, IVs, etc.)
        /// are randomly generated; only the initial password and the iteration
        /// count need to be specified.
        /// </summary>
        /// <param name="password">Initial password to use for the storage</param>
        /// <param name="iterationCount">Iteration count to use when deriving the password key</param>
        /// <param name="controller">Slow operation status</param>
        public EncryptionHeader(
            string password,
            uint iterationCount,
            SlowOperationController controller)
        {
            EncryptDataKey(
                Crypto.GenerateKey(),
                password,
                iterationCount,
                controller);
        }

        /// <summary>
        /// Creates a Content object to represent the header's contents
        /// </summary>
        /// <returns>Header's contents</returns>
        public Content ToContent()
        {
            MemoryStream headerStream = new MemoryStream();

            Utils.WriteByteArrayWithLengthToStream(
                passwordSalt,
                headerStream,
                false/*estimate*/);
            Utils.WriteUIntToStream(
                iterationCount,
                headerStream,
                false/*estimate*/);
            Utils.WriteByteArrayWithLengthToStream(
                passwordVerifier,
                headerStream,
                false/*estimate*/);
            Utils.WriteByteArrayWithLengthToStream(
                encryptedDataKey,
                headerStream,
                false/*estimate*/);
            Utils.WriteByteArrayWithLengthToStream(
                dataKeyIV,
                headerStream,
                false/*estimate*/);

            return new Content(headerStream);
        }

        /// <summary>
        /// Returns the decrypted data key, if the password supplied is correct
        /// </summary>
        /// <param name="password">Password to unlock the data key</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>Decrypted data key</returns>
        public DataKey GetDataKey(
            string password,
            SlowOperationController controller)
        {
            controller.StartProgress(2);

            try
            {
                // Derive a key from the password
                DataKey passwordKey
                    = Crypto.KeyFromPassword(
                        password,
                        passwordSalt,
                        iterationCount,
                        controller);

                controller.ReportProgress(1);

                // And verify that the password is correct
                if (!Crypto.VerifyPassword(
                        passwordKey,
                        passwordVerifier,
                        controller))
                {
                    // Wrong password!
                    throw new WrongPasswordException(); // TODO: Fix exceptions
                }

                controller.ReportProgress(2);

                // If so, use the password key to decrypt the data key
                {
                    Content decryptedKeyContent
                        = Crypto.Decrypt(
                            new EncryptedContent(
                                new MemoryStream(encryptedDataKey),
                                new IV(dataKeyIV)),
                            passwordKey);

                    DataKey key = new DataKey(decryptedKeyContent.DetachData());
                    
                    decryptedKeyContent.Dispose();
                    return key;
                }
            }
            finally
            {
                // Every StartProgress must have a corresponding FinishProgress!
                controller.FinishProgress();
            }
        }

        /// <summary>
        /// Changes the user's password in the encryption header
        /// </summary>
        /// <param name="oldPassword">Old password</param>
        /// <param name="newPassword">New password</param>
        /// <param name="newIterationCount">New iteration count to use</param>
        /// <param name="controller">Slow operation controller</param>
        public void ChangePassword(
            string oldPassword,
            string newPassword,
            uint newIterationCount,
            SlowOperationController controller)
        {
            controller.StartProgress(2);
            DataKey key = GetDataKey(oldPassword, controller);
            controller.ReportProgress(1);
            EncryptDataKey(key, newPassword, newIterationCount, controller);
            controller.ReportProgress(2);
            controller.FinishProgress();
        }

        private void EncryptDataKey(
            DataKey dataKey,
            string password,
            uint iterationCount,
            SlowOperationController controller)
        {
            passwordSalt = Crypto.GenerateSalt();
            this.iterationCount = iterationCount;

            controller.StartProgress(2);
            try
            {
                DataKey passwordKey
                    = Crypto.KeyFromPassword(
                        password,
                        passwordSalt,
                        iterationCount,
                        controller);

                controller.ReportProgress(1);

                passwordVerifier
                    = Crypto.VerifierFromPassword(
                        passwordKey,
                        controller);

                controller.ReportProgress(2);

                // Encrypt the data key with the password key
                {
                    Content dataKeyContent = null;
                    EncryptedContent encryptedDataKey = null;
                    Stream encryptedDataKeyStream = null;

                    try
                    {
                        dataKeyContent = new Content(dataKey.Bytes);
                        encryptedDataKey = Crypto.Encrypt(dataKeyContent, passwordKey);
                        encryptedDataKeyStream = encryptedDataKey.DetachData();

                        this.encryptedDataKey = Utils.ReadStream(encryptedDataKeyStream);
                        dataKeyIV = encryptedDataKey.IV.Bytes;
                    }
                    finally
                    {
                        if (encryptedDataKeyStream != null)
                        {
                            encryptedDataKeyStream.Close();
                        }

                        if (encryptedDataKey != null)
                        {
                            encryptedDataKey.Dispose();
                        }

                        if (dataKeyContent != null)
                        {
                            dataKeyContent.Dispose();
                        }
                    }
                }
            }
            finally
            {
                controller.FinishProgress();
            }
        }

        /// <summary>
        /// Salt to use with PBKDF2 to derive a key from the password
        /// </summary>
        private byte[] passwordSalt;

        /// <summary>
        /// Number of iterations the PBKDF2 algorithm needs to make when
        /// deriving a key from the password
        /// </summary>
        private uint iterationCount;

        /// <summary>
        /// Password verifier - used to verify whether the password supplied by
        /// the user is correct (otherwise we wouldn't know and may try to
        /// decrypt with the wrong key, which will result in garbage and other
        /// unpleasant behavior).
        /// </summary>
        private byte[] passwordVerifier;

        /// <summary>
        /// Encrypted master key for access to the private data. It can be
        /// decrypted with the key derived from the user's password.
        /// </summary>
        private byte[] encryptedDataKey;
        private byte[] dataKeyIV;
    }
}
