using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Reflection;

using PrivateDataCommon;

namespace PrivateDataCore
{
    /// <summary>
    /// Possible item types Private Data can handle
    /// </summary>
    public enum ItemType
    {
        Folder,
        Note,
        GenericPassword,
        WebAccount,
        Email,
        BankAccount,
        CreditCard,
        CellPhone,
        Picture,
        File
    }

    /// <summary>
    /// Base class for all Private Data Item implementations
    /// </summary>
    public abstract class Item : IDisposable
    {
        static Item()
        {
            itemConstructors = new Dictionary<ItemType, ItemConstructorInfo>();

            itemConstructors.Add(
                ItemType.Folder,
                GetItemConstructorInfo(typeof(FolderItem)));
            itemConstructors.Add(
                ItemType.Note,
                GetItemConstructorInfo(typeof(ItemTypes.Note)));
            itemConstructors.Add(
                ItemType.GenericPassword,
                GetItemConstructorInfo(typeof(ItemTypes.GenericPassword)));
            itemConstructors.Add(
                ItemType.WebAccount,
                GetItemConstructorInfo(typeof(ItemTypes.WebAccount)));
            itemConstructors.Add(
                ItemType.Email,
                GetItemConstructorInfo(typeof(ItemTypes.Email)));
            itemConstructors.Add(
                ItemType.BankAccount,
                GetItemConstructorInfo(typeof(ItemTypes.BankAccount)));
            itemConstructors.Add(
                ItemType.CreditCard,
                GetItemConstructorInfo(typeof(ItemTypes.CreditCard)));
            itemConstructors.Add(
                ItemType.CellPhone,
                GetItemConstructorInfo(typeof(ItemTypes.CellPhone)));
            itemConstructors.Add(
                ItemType.Picture,
                GetItemConstructorInfo(typeof(ItemTypes.Picture)));
            itemConstructors.Add(
                ItemType.File,
                GetItemConstructorInfo(typeof(ItemTypes.FileItem)));
        }

        private static ItemConstructorInfo GetItemConstructorInfo(Type type)
        {
            ConstructorInfo createConstructor = null;
            ConstructorInfo loadConstructor = null;

            // Get the type's non-public instance constructors
            ConstructorInfo[] constructors
                = type.GetConstructors(
                    BindingFlags.NonPublic | BindingFlags.Instance);

            // Iterate through the constructors and find the ones we want
            foreach (ConstructorInfo constructor in constructors)
            {
                ParameterInfo[] parameters = constructor.GetParameters();

                if (parameters.Length == 1
                    && parameters[0].ParameterType == typeof(Metadata))
                {
                    createConstructor = constructor;
                }

                if (parameters.Length == 3
                    && parameters[0].ParameterType == typeof(ItemID)
                    && parameters[1].ParameterType == typeof(Metadata)
                    && parameters[2].ParameterType == typeof(Content))
                {
                    loadConstructor = constructor;
                }
            }

            // Verify that we found both constructors we were interested in
            if (createConstructor == null || loadConstructor == null)
            {
                // Internal error: we couldn't find one or both constructors
                throw new Exception(); // TODO: Fix exceptions
            }

            return new ItemConstructorInfo(createConstructor, loadConstructor);
        }

        public static Item Create(Metadata metadata)
        {
            return (Item)itemConstructors[metadata.Type].CreateConstructor.Invoke(
                new object[] { metadata });
        }

        public static Item Create(ItemID id, Metadata metadata, Content content)
        {
            return (Item)itemConstructors[metadata.Type].LoadConstructor.Invoke(
                new object[] { id, metadata, content });
        }

        protected Item(Metadata metadata)
        {
            this.id = new ItemID();
            this.metadata = metadata;
        }

        protected Item(ItemID id, Metadata metadata)
        {
            this.id = id;
            this.metadata = metadata;
        }

        public ItemID ID
        {
            get
            {
                return id;
            }

            set
            {
                if (id != null)
                {
                    // No change of ID is allowed; only assigning a new ID
                    throw new Exception(); // TODO: Fix exceptions
                }

                id = value;
            }
        }

        public string Name
        {
            get
            {
                return metadata.Name;
            }

            set
            {
                metadata.Name = value;
            }
        }

        public ItemType Type
        {
            get
            {
                return metadata.Type;
            }
        }

        public string FileName
        {
            get
            {
                return metadata.FileName;
            }

            set
            {
                metadata.FileName = value;
            }
        }

        public string Notes
        {
            get
            {
                return notes;
            }

            set
            {
                notes = value;
            }
        }

        public Content GenerateMetadata()
        {
            return metadata.ToContent();
        }

        public abstract Content GenerateContent();

        private struct ItemConstructorInfo
        {
            public ItemConstructorInfo(
                ConstructorInfo createConstructor,
                ConstructorInfo loadConstructor)
            {
                this.createConstructor = createConstructor;
                this.loadConstructor = loadConstructor;
            }

            public ConstructorInfo CreateConstructor
            {
                get
                {
                    return createConstructor;
                }
            }

            public ConstructorInfo LoadConstructor
            {
                get
                {
                    return loadConstructor;
                }
            }

            private ConstructorInfo createConstructor;
            private ConstructorInfo loadConstructor;
        }

        /// <summary>
        /// Disposes the resources held by the object (inherited from IDisposable)
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
        }

        private static Dictionary<ItemType, ItemConstructorInfo> itemConstructors;

        private ItemID id;
        private Metadata metadata;

        protected string notes = String.Empty;
    }
}
