using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;

using PrivateDataCommon;

namespace PrivateDataCore
{
    internal class ItemTree
    {
        public ItemTree(DataManager dataManager, ItemID[] items)
        {
            this.dataManager = dataManager;
            parent = new Dictionary<ItemID, ItemID>(items.Length);
            root = new List<ItemID>();

            // 1. Fill the parent list, all items being under the root
            foreach (ItemID id in items)
            {
                parent.Add(id, null);
            }

            // 2. Go over all folder items and update the parent list
            foreach (ItemID id in items)
            {
                Metadata metadata = dataManager.GetMetadata(id);

                if (metadata.Type == ItemType.Folder)
                {
                    FolderItem folder = dataManager.GetItem(id) as FolderItem;

                    foreach (ItemID child in folder.Items)
                    {
                        parent[child] = id;
                    }

                    // TODO: Dispose folder
                }
            }

            // 3. Assemble a list of items still under the root
            foreach (ItemID id in items)
            {
                if (parent[id] == null)
                {
                    root.Add(id);
                }
            }
        }

        public ItemID[] GetFolderItems(ItemID folder)
        {
            // Was the request for the root?
            if (folder == null)
            {
                return root.ToArray();
            }

            FolderItem folderItem = dataManager.GetItem(folder) as FolderItem;
            if (folderItem == null)
            {
                // Request for a non-existent folder
                throw new Exception(); // TODO: Fix exceptions
            }

            try
            {
                return folderItem.Items.ToArray();
            }
            finally
            {
                // TODO: Dispose folderItem
            }
        }

        /// <summary>
        /// Moves an item into a specified folder (or puts a new item into a
        /// specified folder)
        /// </summary>
        /// <param name="item">Item to file into the folder</param>
        /// <param name="folder">Folder to file the item into</param>
        /// <param name="rootChanged">
        /// Informs the caller whether the contents of the root folder changed.
        /// The caller may need this information in order to be able to update
        /// the user interface properly. Unlike other folders, for which we
        /// call DataManager.WriteItem, which can update the user interface as
        /// appropriate, we do not call the data manager when updating the root
        /// folder. Therefore, we need to return this information explicitly.
        /// </param>
        /// <param name="controller">Slow operation controller</param>
        public void MoveItemToFolder(
            ItemID item,
            ItemID folder,
            out bool rootChanged,
            SlowOperationController controller)
        {
            // Unless determined otherwise, we did not modify the root
            rootChanged = false;

            // Verify that the folder is actually a folder
            if (folder != null)
            {
                Metadata folderMetadata = dataManager.GetMetadata(folder);
                if (folderMetadata.Type != ItemType.Folder)
                {
                    // Why did the user interface allow this?
                    throw new Exception(); // TODO: Fix exceptions
                }
            }

            // First, let's clear the item from the source folder. This way we
            // guarantee that, should an error occur, we won't end up with the
            // item being in both folders
            {
                ItemID source;
                if (parent.TryGetValue(item, out source))
                {
                    // If the item is already where it should be, abort now.
                    // This also prevents an avalanche of folder updates
                    // (since every item, when written to the storage, is filed
                    // into its appropriate folder no matter whether already in
                    // it or not, rewriting the item's folder will try to "put"
                    // the folder itself under its parent again).
                    if (source == folder)
                    {
                        return;
                    }

                    parent[item] = null;
                    if (source != null)
                    {
                        FolderItem sourceFolder = (FolderItem)dataManager.GetItem(source);
                        sourceFolder.Items.Remove(item);
                        dataManager.StoreItem(
                            sourceFolder,
                            parent[sourceFolder.ID],
                            controller);
                        // TODO: Dispose sourceFolder
                    }
                    else
                    {
                        root.Remove(item);
                        rootChanged = true;
                    }
                }
            }

            // And put it in the destination folder
            parent[item] = folder;
            if (folder != null)
            {
                FolderItem destinationFolder = (FolderItem)dataManager.GetItem(folder);
                destinationFolder.Items.Add(item);
                dataManager.StoreItem(
                    destinationFolder,
                    parent[destinationFolder.ID],
                    controller);
                // TODO: Dispose destinationFolder
            }
            else
            {
                root.Add(item);
                rootChanged = true;
            }
        }

        public void DeleteItem(ItemID item, SlowOperationController controller)
        {
            ItemID itemParent = parent[item];
            parent.Remove(item);

            if (itemParent != null)
            {
                FolderItem parentFolder = (FolderItem)dataManager.GetItem(itemParent);
                parentFolder.Items.Remove(item);
                dataManager.StoreItem(
                    parentFolder,
                    parent[parentFolder.ID],
                    controller);
                // TODO: Dispose sourceFolder
            }
            else
            {
                if (!root.Remove(item))
                {
                    // We didn't find or were unable to remove the item
                    throw new Exception(); // TODO: Fix exceptions
                }
            }
        }

        private DataManager dataManager;

        private Dictionary<ItemID, ItemID> parent;
        private List<ItemID> root;
    }
}
