using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Represents a single custom data item
    /// </summary>
    internal class CustomDataItem : IDisposable
    {
        /// <summary>
        /// Constructs a custom data item object
        /// </summary>
        public CustomDataItem(DataFileManager.GenericDataBlockDescriptor content,
                              DataFileManager.GenericDataBlockDescriptor contentIV)
        {
            contentDescriptor = content;
            contentIVDescriptor = contentIV;
        }

        /// <summary>
        /// Reads the custom data item record from a stream
        /// </summary>
        /// <param name="fileManager">Data file manager</param>
        /// <param name="stream">Stream to read from</param>
        public CustomDataItem(DataFileManager fileManager, Stream stream)
        {
            contentDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(
                    fileManager,
                    stream);

            contentIVDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(
                    fileManager,
                    stream);
        }

        /// <summary>
        /// Destroys the object
        /// </summary>
        ~CustomDataItem()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// Disposes the resources held by the object (inherited from IDisposable)
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                contentDescriptor.Dispose();
                contentIVDescriptor.Dispose();
            }
#if DEBUG
            else
            {
                // Somebody forgot to call Dispose explicitly!!!
                throw new Exception(); // TODO: Fix exceptions
            }
#endif
        }

        /// <summary>
        /// Writes the custom data item record to a stream
        /// </summary>
        /// <param name="stream">Stream to write to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public uint WriteToStream(Stream stream, bool estimate)
        {
            uint bytesWritten = 0;

            bytesWritten += contentDescriptor.WriteToStream(stream, estimate);
            bytesWritten += contentIVDescriptor.WriteToStream(stream, estimate);

            // Keep file in consistent state
            if (!estimate)
            {
                stream.Flush();
            }

            return bytesWritten;
        }

        /// <summary>
        /// Gets the custom data item's content. If the item is encrypted,
        /// the returned object is of type EncryptedContent.
        /// </summary>
        public Content Content
        {
            get
            {
                if (contentDescriptor.BlockLength < 0)
                {
                    // Invalid content length
                    throw new Exception(); // TODO: Fix exceptions
                }

                MemoryStream content
                    = new MemoryStream(contentDescriptor.ReadData());

                return
                    contentIVDescriptor.BlockLength > 0
                        ? new EncryptedContent(
                            content,
                            new IV(contentIVDescriptor.ReadData()))
                        : new Content(content);
            }
        }

        private DataFileManager.GenericDataBlockDescriptor contentDescriptor;
        private DataFileManager.GenericDataBlockDescriptor contentIVDescriptor;
    }
}
