using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Array of custom data items
    /// </summary>
    internal class CustomDataItemArray : IDisposable
    {
        /// <summary>
        /// Element of the custom data item array. Generally this would be
        /// only the descriptor to the custom data item, but we need to
        /// store the item object as well so the descriptors it holds are
        /// around to keep the data they point to marked as not free.
        /// </summary>
        private class CustomDataItemArrayElement : IDisposable
        {
            /// <summary>
            /// 
            /// </summary>
            /// <param name="fileManager"></param>
            public CustomDataItemArrayElement(DataFileManager fileManager)
            {
                descriptor
                    = new DataFileManager.CustomDataItemDataBlockDescriptor(
                        fileManager);
                item = null;
            }

            /// <summary>
            /// Reads the custom data item array element from a stream
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Stream to read from</param>
            public CustomDataItemArrayElement(DataFileManager fileManager, Stream stream)
            {
                // Read the descriptor from the stream
                descriptor
                    = new DataFileManager.CustomDataItemDataBlockDescriptor(
                        fileManager,
                        stream);

                // Read the item as well but be careful to leave the stream
                // position intact
                if (descriptor.BlockLength > 0)
                {
                    long offset = stream.Position;
                    item = descriptor.ReadData();
                    stream.Seek(offset, SeekOrigin.Begin);
                }
            }

            /// <summary>
            /// Destroys the object
            /// </summary>
            ~CustomDataItemArrayElement()
            {
                Dispose(false/*disposing*/);
            }

            /// <summary>
            /// Disposes the resources held by the object (inherited from IDisposable)
            /// </summary>
            public void Dispose()
            {
                Dispose(true/*disposing*/);
                GC.SuppressFinalize(this);
            }

            /// <summary>
            /// Disposes the resources held by the object
            /// </summary>
            /// <param name="disposing">
            /// True if this method is called directly from user code. False if
            /// Dispose has been called by the runtime through the class' finalizer
            /// (in this case, we can't dereference / try to dispose resources
            /// held by the object as they may have already been destroyed).
            /// </param>
            protected virtual void Dispose(bool disposing)
            {
                if (disposing)
                {
                    if (item != null)
                    {
                        item.Dispose();
                        item = null;
                    }

                    if (descriptor != null)
                    {
                        descriptor.Dispose();
                        descriptor = null;
                    }
                }
#if DEBUG
                else
                {
                    // Somebody forgot to call Dispose explicitly!!!
                    throw new Exception(); // TODO: Fix exceptions
                }
#endif
            }

            /// <summary>
            /// 
            /// </summary>
            public CustomDataItem Item
            {
                get
                {
                    return item;
                }

                set
                {
                    item = value;
                }
            }

            /// <summary>
            /// 
            /// </summary>
            public DataFileManager.CustomDataItemDataBlockDescriptor Descriptor
            {
                get
                {
                    return descriptor;
                }
            }

            private CustomDataItem item;
            private DataFileManager.CustomDataItemDataBlockDescriptor descriptor;
        }

        /// <summary>
        /// Constructs an empty custom data item array
        /// </summary>
        /// <param name="fileManager">Data file manager</param>
        public CustomDataItemArray(DataFileManager fileManager)
        {
            this.fileManager = fileManager;
            items = new List<CustomDataItemArrayElement>();
        }

        /// <summary>
        /// Reads the custom data item array from a stream
        /// </summary>
        /// <param name="fileManager">Data file manager</param>
        /// <param name="stream">Stream to read from</param>
        public CustomDataItemArray(DataFileManager fileManager, Stream stream)
        {
            int customDataItemCount = Utils.ReadIntFromStream(stream);

            this.fileManager = fileManager;
            items
                = new List<CustomDataItemArrayElement>(customDataItemCount);

            for (int i = 0; i < customDataItemCount; i++)
            {
                items.Add(new CustomDataItemArrayElement(fileManager, stream));
            }
        }

        /// <summary>
        /// Destroys the object
        /// </summary>
        ~CustomDataItemArray()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// Disposes the resources held by the object (inherited from IDisposable)
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (items != null)
                {
                    for (int i = 0; i < items.Count; i++)
                    {
                        items[i].Dispose();
                    }
                    items = null;
                }
            }
#if DEBUG
            else
            {
                // Somebody forgot to call Dispose explicitly!!!
                throw new Exception(); // TODO: Fix exceptions
            }
#endif
        }

        /// <summary>
        /// Writes the custom data item array to a stream
        /// </summary>
        /// <param name="stream">Stream to write to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public uint WriteToStream(Stream stream, bool estimate)
        {
            uint bytesWritten = 0;

            // Write the number of custom data types out
            bytesWritten += Utils.WriteIntToStream(items.Count, stream, estimate);

            // Write the descriptors themselves
            for (int i = 0; i < items.Count; i++)
            {
                bytesWritten
                    += items[i].Descriptor.WriteToStream(stream, estimate);
            }

            // Keep file in consistent state
            if (!estimate)
            {
                stream.Flush();
            }

            return bytesWritten;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="type"></param>
        /// <param name="item"></param>
        public void Store(int type, CustomDataItem item)
        {
            // If the array is not big enough, we'll need to expand it and
            // reallocate the necessary space in the data file
            if (items.Count <= type)
            {
                // Expand the array to the needed size
                for (int i = items.Count; i <= type; i++)
                {
                    items.Add(new CustomDataItemArrayElement(fileManager));
                }

                // Write the newly expanded array to a new location in the
                // data file
                DataFileManager.CustomDataItemArrayDataBlockDescriptor
                    newArrayDescriptor
                        = new DataFileManager.CustomDataItemArrayDataBlockDescriptor(
                            fileManager);
                newArrayDescriptor.WriteData(this);

                // Repoint the custom data array descriptor in the file
                // header to the newly written array
                fileManager.FileHeader.CustomDataItemArrayDescriptor.AssignFrom(
                    newArrayDescriptor);

                // We have already assigned this descriptor away, so we no
                // longer need it
                newArrayDescriptor.Dispose();
            }

            // Write the custom data item to the file
            DataFileManager.CustomDataItemDataBlockDescriptor itemDescriptor
                = new DataFileManager.CustomDataItemDataBlockDescriptor(
                    fileManager);
            itemDescriptor.WriteData(item);

            // And repoint the respective descriptor in the array to it
            items[type].Descriptor.AssignFrom(itemDescriptor);
            if (items[type].Item != null)
            {
                // TODO: Can we get the Item set accessor to do this?
                items[type].Item.Dispose();
            }
            items[type].Item = item;

            // We have already assigned this descriptor away, so we no
            // longer need it
            itemDescriptor.Dispose();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public CustomDataItem this[int type]
        {
            get
            {
                return items[type].Item;
            }
        }

        private DataFileManager fileManager;
        private List<CustomDataItemArrayElement> items;
    }
}
