using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Manages data block and their descriptors in the main data file
    /// </summary>
    internal class DataFileManager : IDisposable
    {
        /// <summary>
        /// Represents a descriptor of a main data file's data block and
        /// offers methods for storing and retrieving the data block's data.
        /// </summary>
        public abstract class DataBlockDescriptor : IDisposable
        {
            /// <summary>
            /// Initializes a new data block descriptor
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            public DataBlockDescriptor(DataFileManager fileManager)
            {
                this.fileManager = fileManager;
            }

            /// <summary>
            /// Reads a data block descriptor from a stream and registers
            /// it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Stream to read from</param>
            public DataBlockDescriptor(DataFileManager fileManager,
                                       Stream stream)
            {
                this.fileManager = fileManager;
                descriptorStream = stream;
                descriptorOffset = (ulong)stream.Position;

                // Read the descriptor from the stream
                ReadFromStream(stream);

                // Now that we have data, it's time to register with the
                // file manager
                if (dataBlockStream != null)
                {
                    fileManager.RegisterDescriptor(this);
                }
            }

            /// <summary>
            /// Destroys the object
            /// </summary>
            ~DataBlockDescriptor()
            {
                Dispose(false/*disposing*/);
            }

            /// <summary>
            /// Disposes the resources held by the object (inherited from IDisposable)
            /// </summary>
            public void Dispose()
            {
                Dispose(true/*disposing*/);
                GC.SuppressFinalize(this);
            }

            /// <summary>
            /// Disposes the resources held by the object
            /// </summary>
            /// <param name="disposing">
            /// True if this method is called directly from user code. False if
            /// Dispose has been called by the runtime through the class' finalizer
            /// (in this case, we can't dereference / try to dispose resources
            /// held by the object as they may have already been destroyed).
            /// </param>
            protected virtual void Dispose(bool disposing)
            {
                if (disposing)
                {
                    if (fileManager != null)
                    {
                        // If we have data, then we registered with the
                        // file manager, so we need to unregister
                        if (dataBlockStream != null)
                        {
                            fileManager.RemoveDescriptor(this);
                        }

                        fileManager = null;
                    }
                }
#if DEBUG
                else
                {
                    // Somebody forgot to call Dispose explicitly!!!
                    throw new Exception(); // TODO: Fix exceptions
                }
#endif
            }

            /// <summary>
            /// "Re-points" the descriptor to the same data block as the
            /// one described by a source descriptor. After executing this,
            /// both descriptors describe the same thing. Notice that the
            /// block types must match!
            /// 
            /// NOTE: This method writes the newly changed descriptor back
            ///       to the file, so the file's position may be different
            ///       after this method's execution.
            /// </summary>
            /// <param name="source">
            /// Source descriptor. It should be of the same type as the
            /// destination.
            /// </param>
            public virtual void AssignFrom(DataBlockDescriptor source)
            {
                if (source.GetType() != this.GetType())
                {
                    // The data blocks should be of the same type
                    throw new Exception(); // TODO: Fix exceptions
                }

                // If we have data, then we registered with the
                // file manager, so we need to unregister
                if (dataBlockStream != null)
                {
                    fileManager.RemoveDescriptor(this);
                }

                dataBlockStream = source.dataBlockStream;
                dataBlockOffset = source.dataBlockOffset;
                dataBlockLength = source.dataBlockLength;

                // We have new data - the descriptor needs to be re-saved
                if (descriptorStream != null)
                {
                    descriptorStream.Seek((long)descriptorOffset, SeekOrigin.Begin);
                    WriteToStream(descriptorStream, false/*estimate*/);
                }

                // If we have new data - re-register with the file manager
                if (dataBlockStream != null)
                {
                    fileManager.RegisterDescriptor(this);
                }
            }

            /// <summary>
            /// Writes the data block descriptor to a stream.
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            public abstract uint WriteToStream(Stream stream, bool estimate);

            /// <summary>
            /// Reads a data block descriptor from a stream.
            /// Used during construction.
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected abstract void ReadFromStream(Stream stream);

            /// <summary>
            /// Writes the offset of the data block to a stream
            /// 
            /// NOTE: The offset must be the very first element of a data
            ///       block descriptor!!!
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            protected uint WriteBlockOffsetToStream(Stream stream, bool estimate)
            {
                // Remember where the descriptor is in the file so that
                // it can be re-saved later when modified
                if (!estimate)
                {
                    descriptorStream = stream;
                    descriptorOffset = (ulong)stream.Position;
                }

                // Write the offset
                return Utils.WriteULongToStream(dataBlockOffset, stream, estimate);
            }

            /// <summary>
            /// Reads the offset of the data block from a stream
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected void ReadBlockOffsetFromStream(Stream stream)
            {
                dataBlockOffset = Utils.ReadULongFromStream(stream);
                dataBlockStream = (dataBlockOffset > 0 ? stream : null);
            }

            /// <summary>
            /// Writes the length of the data block to a stream
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            protected uint WriteBlockLengthToStream(Stream stream, bool estimate)
            {
                // Write the length
                return Utils.WriteUIntToStream(dataBlockLength, stream, estimate);
            }

            /// <summary>
            /// Reads the length of the data block from a stream
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected void ReadBlockLengthFromStream(Stream stream)
            {
                dataBlockLength = Utils.ReadUIntFromStream(stream);
            }

            /// <summary>
            /// Sets the data block's length to a pre-defined known value
            /// (used by derived classes whose data blocks have pre-defined
            /// lengths that need not be written out)
            /// </summary>
            /// <param name="length">Length of the data block</param>
            protected void SetPredefinedBlockLength(uint length)
            {
                dataBlockLength = length;
            }

            /// <summary>
            /// Allocates space for the data block on the stream
            /// </summary>
            /// <param name="length">Length of the data block</param>
            /// <returns>Stream ready to accept the data block</returns>
            protected Stream AllocateBlock(uint length)
            {
                dataBlockOffset = fileManager.AllocateDataBlock(length);
                dataBlockLength = length;
                dataBlockStream = fileManager.stream;

                // Now that we have allocated the block, it's time to
                // register with the file manager
                fileManager.RegisterDescriptor(this);

                dataBlockStream.Seek((long)dataBlockOffset, SeekOrigin.Begin);
                return dataBlockStream;
            }

            /// <summary>
            /// Prepares to read the block from the stream
            /// </summary>
            /// <param name="stream">Stream to read the data from</param>
            /// <param name="length">Length of the data block</param>
            protected void GetBlock(out Stream stream, out uint length)
            {
                dataBlockStream.Seek((long)dataBlockOffset, SeekOrigin.Begin);

                stream = dataBlockStream;
                length = dataBlockLength;
            }

            /// <summary>
            /// Moves the data block to a new location in the data file.
            /// Used during defragmentation.
            /// </summary>
            /// <param name="newOffset">New offset</param>
            /// <param name="controller">Slow operation controller</param>
            /// <returns>Number of bytes moved</returns>
            public uint MoveBlockTo(ulong newOffset, SlowOperationController controller)
            {
                // Copy the data
                Utils.CopyInsideStream(
                    dataBlockStream,
                    dataBlockOffset,
                    newOffset,
                    dataBlockLength,
                    controller);

                // Unregister the descriptor so it gets re-filed correctly
                // when we're done
                if (dataBlockStream != null)
                {
                    fileManager.RemoveDescriptor(this);
                }

                dataBlockOffset = newOffset;

                // We have a new offset - the descriptor needs to be re-saved
                if (descriptorStream != null)
                {
                    descriptorStream.Seek((long)descriptorOffset, SeekOrigin.Begin);
                    WriteToStream(descriptorStream, false/*estimate*/);
                }

                // We have a new offset - re-register with the file manager
                if (dataBlockStream != null)
                {
                    fileManager.RegisterDescriptor(this);
                }

                return dataBlockLength;
            }

            // TODO: Don't call on any block move; notify only if the descriptor has moved
            /// <summary>
            /// Notifies the descriptor that it may have been moved (for
            /// example during file defragmentation) so that it knows its
            /// new location in the file
            /// </summary>
            /// <param name="offset">Offset of the moved data block</param>
            /// <param name="length">Length of the moved data block</param>
            /// <param name="newOffset">New offset of the data block</param>
            public void OnMove(ulong offset, uint length, ulong newOffset)
            {
                if (descriptorOffset >= offset
                        && descriptorOffset < offset + length)
                {
                    descriptorOffset += (newOffset - offset);
                }
            }

            /// <summary>
            /// Offset of the data block in the file. DO NOT USE!!!
            /// This property should only be used by the Space Manager / 
            /// defragmenter so that they can know where data is. Everyone
            /// else should just use ReadBlock / ReadData (in derived classes).
            /// </summary>
            public ulong BlockOffset
            {
                get
                {
                    return dataBlockOffset;
                }
            }

            /// <summary>
            /// Length of the data block
            /// </summary>
            public uint BlockLength
            {
                get
                {
                    return dataBlockLength;
                }
            }

            // Data file manager
            protected DataFileManager fileManager;

            // Position of the descriptor itself
            protected Stream descriptorStream;
            protected ulong descriptorOffset;

            // Descriptor data
            private Stream dataBlockStream;
            private ulong dataBlockOffset;
            private uint dataBlockLength;
        }

        /// <summary>
        /// Descriptor of a generic data block
        /// </summary>
        public class GenericDataBlockDescriptor : DataBlockDescriptor
        {
            /// <summary>
            /// Initializes a new generic data block descriptor and
            /// registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            public GenericDataBlockDescriptor(DataFileManager fileManager)
                : base(fileManager)
            {
            }

            /// <summary>
            /// Reads a generic data block descriptor from a stream and
            /// registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Stream to read from</param>
            public GenericDataBlockDescriptor(DataFileManager fileManager,
                                              Stream stream)
                : base(fileManager, stream)
            {
            }

            /// <summary>
            /// Writes the descriptor to a stream
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            public override uint WriteToStream(Stream stream, bool estimate)
            {
                uint bytesWritten = 0;

                bytesWritten += WriteBlockOffsetToStream(stream, estimate);
                bytesWritten += WriteBlockLengthToStream(stream, estimate);

                // Keep file in consistent state
                if (!estimate)
                {
                    stream.Flush();
                }

                return bytesWritten;
            }

            /// <summary>
            /// Reads the descriptor from a stream
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected override void ReadFromStream(Stream stream)
            {
                ReadBlockOffsetFromStream(stream);
                ReadBlockLengthFromStream(stream);
            }

            /// <summary>
            /// Writes the data block's contents
            /// </summary>
            /// <param name="data">
            /// Data block's contents. If null, the data is considered to
            /// have been saved externally.
            /// </param>
            public void WriteData(byte[] data)
            {
                if (data != null)
                {
                    Stream stream = AllocateBlock((uint)data.Length);
                    Utils.WriteByteArrayToStream(data, stream);
                }
                else
                {
                    SetPredefinedBlockLength(externalLength);
                }
            }

            /// <summary>
            /// Reads the data block's contents
            /// </summary>
            /// <returns>Contents of the data block</returns>
            public byte[] ReadData()
            {
                if (BlockLength != externalLength)
                {
                    Stream stream;
                    uint length;
                    GetBlock(out stream, out length);

                    byte[] data = new byte[length];
                    Utils.FillByteArrayFromStream(data, stream);
                    return data;
                }
                else
                {
                    return null;
                }
            }
        }

        /// <summary>
        /// Descriptor of an item list element's data block
        /// </summary>
        public class ListElementDataBlockDescriptor : DataBlockDescriptor
        {
            /// <summary>
            /// Initializes a new list element data block descriptor and
            /// registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            public ListElementDataBlockDescriptor(DataFileManager fileManager)
                : base(fileManager)
            {
            }

            /// <summary>
            /// Reads a list element data block descriptor from a stream
            /// and registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Stream to read from</param>
            public ListElementDataBlockDescriptor(DataFileManager fileManager,
                                                  Stream stream)
                : base(fileManager, stream)
            {
            }

            /// <summary>
            /// Writes the descriptor to a stream
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            public override uint WriteToStream(Stream stream, bool estimate)
            {
                uint bytesWritten = 0;

                bytesWritten += WriteBlockOffsetToStream(stream, estimate);

                // Keep file in consistent state
                if (!estimate)
                {
                    stream.Flush();
                }

                return bytesWritten;
            }

            /// <summary>
            /// Reads the descriptor from a stream
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected override void ReadFromStream(Stream stream)
            {
                // Calculate the length of a list element, if we still
                // haven't done so (we do this only once)
                if (listElementLength == 0)
                {
                    ItemList.ListElement listElement
                        = new ItemList.ListElement(
                            new ListElementDataBlockDescriptor(fileManager),
                            new ItemID(),
                            new Item(
                                new GenericDataBlockDescriptor(fileManager),
                                new GenericDataBlockDescriptor(fileManager),
                                new GenericDataBlockDescriptor(fileManager),
                                new GenericDataBlockDescriptor(fileManager)));

                    listElementLength
                        = listElement.WriteToStream(null/*stream*/, true/*estimate*/);

                    listElement.Dispose();
                }

                ReadBlockOffsetFromStream(stream);
                if (BlockOffset > 0)
                {
                    SetPredefinedBlockLength(listElementLength);
                }
                else
                {
                    // Null descriptor should have empty length
                    SetPredefinedBlockLength(0);
                }
            }

            /// <summary>
            /// Writes the data block's contents out
            /// </summary>
            /// <param name="data">Data block's contents</param>
            public void WriteData(ItemList.ListElement listElement)
            {
                // Estimate how much length we need for the list element
                uint length = listElement.WriteToStream(null, true/*estimate*/);

                // Allocate the necessary space
                Stream stream = AllocateBlock(length);

                // Write the list element out
                uint bytesWritten
                    = listElement.WriteToStream(stream, false/*estimate*/);

                if (bytesWritten != length)
                {
                    // Internal error: why was the length not predicted correctly?
                    throw new Exception(); // TODO: Fix exceptions
                }
            }

            /// <summary>
            /// Reads the data block's contents
            /// </summary>
            /// <returns>List element</returns>
            public ItemList.ListElement ReadData()
            {
                Stream stream;
                uint length;
                GetBlock(out stream, out length);

                return new ItemList.ListElement(fileManager, stream);
            }

            private static uint listElementLength = 0;
        }

        /// <summary>
        /// Descriptor of custom data item's data block
        /// </summary>
        public class CustomDataItemDataBlockDescriptor : DataBlockDescriptor
        {
            /// <summary>
            /// Initializes a new custom data item data block descriptor
            /// and registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            public CustomDataItemDataBlockDescriptor(
                                            DataFileManager fileManager)
                : base(fileManager)
            {
            }

            /// <summary>
            /// Reads a custom data item data block descriptor from a stream
            /// and registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Stream to read from</param>
            public CustomDataItemDataBlockDescriptor(
                                            DataFileManager fileManager,
                                            Stream stream)
                : base(fileManager, stream)
            {
            }

            /// <summary>
            /// Writes the descriptor to a stream
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            public override uint WriteToStream(Stream stream, bool estimate)
            {
                uint bytesWritten = 0;

                bytesWritten += WriteBlockOffsetToStream(stream, estimate);

                // Keep file in consistent state
                if (!estimate)
                {
                    stream.Flush();
                }

                return bytesWritten;
            }

            /// <summary>
            /// Reads the descriptor from a stream
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected override void ReadFromStream(Stream stream)
            {
                // Calculate the length of a custom data item, if we still
                // haven't done so (we do this only once)
                if (customDataItemLength == 0)
                {
                    CustomDataItem customDataItem
                        = new CustomDataItem(
                            new GenericDataBlockDescriptor(fileManager),
                            new GenericDataBlockDescriptor(fileManager));

                    customDataItemLength
                        = customDataItem.WriteToStream(
                            null/*stream*/,
                            true/*estimate*/);

                    customDataItem.Dispose();
                }

                ReadBlockOffsetFromStream(stream);
                if (BlockOffset > 0)
                {
                    SetPredefinedBlockLength(customDataItemLength);
                }
                else
                {
                    // Null descriptor should have empty length
                    SetPredefinedBlockLength(0);
                }
            }

            /// <summary>
            /// Writes the data block's contents out
            /// </summary>
            /// <param name="customDataItem">Data block's contents</param>
            public void WriteData(CustomDataItem customDataItem)
            {
                // Estimate how much length we need for the custom data item
                uint length = customDataItem.WriteToStream(null, true/*estimate*/);

                // Allocate the necessary space
                Stream stream = AllocateBlock(length);

                // Write the custom data item
                uint bytesWritten
                    = customDataItem.WriteToStream(stream, false/*estimate*/);

                if (bytesWritten != length)
                {
                    // Internal error: why was the length not predicted correctly?
                    throw new Exception(); // TODO: Fix exceptions
                }
            }

            /// <summary>
            /// Reads the data block's contents
            /// </summary>
            /// <returns>Custom data item</returns>
            public CustomDataItem ReadData()
            {
                Stream stream;
                uint length;
                GetBlock(out stream, out length);

                return new CustomDataItem(fileManager, stream);
            }

            private static uint customDataItemLength = 0;
        }


        /// <summary>
        /// Descriptor of custom data item array's data block
        /// </summary>
        public class CustomDataItemArrayDataBlockDescriptor : DataBlockDescriptor
        {
            /// <summary>
            /// Initializes a new custom data item array data block descriptor
            /// and registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            public CustomDataItemArrayDataBlockDescriptor(
                                            DataFileManager fileManager)
                : base(fileManager)
            {
            }

            /// <summary>
            /// Reads a custom data item array data block descriptor from a
            /// stream and registers it with the data file manager
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Stream to read from</param>
            public CustomDataItemArrayDataBlockDescriptor(
                                            DataFileManager fileManager,
                                            Stream stream)
                : base(fileManager, stream)
            {
            }

            /// <summary>
            /// Writes the descriptor to a stream
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            public override uint WriteToStream(Stream stream, bool estimate)
            {
                uint bytesWritten = 0;

                bytesWritten += WriteBlockOffsetToStream(stream, estimate);
                bytesWritten += WriteBlockLengthToStream(stream, estimate);

                // Keep file in consistent state
                if (!estimate)
                {
                    stream.Flush();
                }

                return bytesWritten;
            }

            /// <summary>
            /// Reads the descriptor from a stream
            /// </summary>
            /// <param name="stream">Stream to read from</param>
            protected override void ReadFromStream(Stream stream)
            {
                ReadBlockOffsetFromStream(stream);
                ReadBlockLengthFromStream(stream);
            }

            /// <summary>
            /// Writes the data block's contents out
            /// </summary>
            /// <param name="cystinDataItemArray">Data block's contents</param>
            public void WriteData(CustomDataItemArray customDataItemArray)
            {
                // Estimate how much length we need for the custom data
                // item array
                uint length
                    = customDataItemArray.WriteToStream(
                        null,
                        true/*estimate*/);

                // Allocate the necessary space
                Stream stream = AllocateBlock(length);

                // Write the custom data item
                uint bytesWritten
                    = customDataItemArray.WriteToStream(
                        stream,
                        false/*estimate*/);

                if (bytesWritten != length)
                {
                    // Internal error: why was the length not predicted correctly?
                    throw new Exception(); // TODO: Fix exceptions
                }
            }

            /// <summary>
            /// Reads the data block's contents
            /// </summary>
            /// <returns>Custom data item array</returns>
            public CustomDataItemArray ReadData()
            {
                Stream stream;
                uint length;
                GetBlock(out stream, out length);

                return new CustomDataItemArray(fileManager, stream);
            }
        }

        /// <summary>
        /// File space manager - manages free/used space
        /// </summary>
        private class SpaceManager
        {
            /// <summary>
            /// Free space manager / allocator
            /// </summary>
            private class FreeSpaceManager
            {
                /// <summary>
                /// Represents a block of free space in the file (we already
                /// have data block descriptors to describe allocated blocks).
                /// </summary>
                private class FreeSpaceBlock
                {
                    /// <summary>
                    /// 
                    /// </summary>
                    /// <param name="offset"></param>
                    /// <param name="length"></param>
                    public FreeSpaceBlock(ulong offset, uint length)
                    {
                        this.offset = offset;
                        this.length = length;
                    }

                    /// <summary>
                    /// Splits a block into two separate blocks
                    /// </summary>
                    /// <param name="offset">
                    /// File offset to split at (must be inside the block)
                    /// </param>
                    /// <param name="first">Resulting first piece</param>
                    /// <param name="second">Resulting second piece</param>
                    public void Split(ulong offset,
                                      out FreeSpaceBlock first,
                                      out FreeSpaceBlock second)
                    {
                        if (offset <= this.offset
                                || offset >= this.offset + this.length)
                        {
                            throw new ArgumentException(); // TODO: Fix exceptions
                        }

                        first
                            = new FreeSpaceBlock(
                                this.offset,
                                (uint)(offset - this.offset));
                        second
                            = new FreeSpaceBlock(
                                offset,
                                (uint)(this.offset + this.length - offset));
                    }

                    /// <summary>
                    /// Offset of the block in the file
                    /// </summary>
                    public ulong Offset
                    {
                        get
                        {
                            return offset;
                        }
                    }

                    /// <summary>
                    /// Length of the block in the file
                    /// </summary>
                    public uint Length
                    {
                        get
                        {
                            return length;
                        }
                    }

                    private ulong offset;
                    private uint length;
                }

                /// <summary>
                /// Creates a new free block list, starting with some
                /// initial free space
                /// </summary>
                /// <param name="offset">Initial free space offset</param>
                /// <param name="length">Initial free space length</param>
                public FreeSpaceManager(ulong offset, uint length)
                {
                    blocks = new LinkedList<FreeSpaceBlock>();
                    if (length > 0)
                    {
                        blocks.AddFirst(
                            new FreeSpaceBlock(offset, length));
                    }

                    freeTotal = length;
                }

                /// <summary>
                /// Allocates (removes from the free space list) and returns
                /// a big enough chunk of free space, if it exists
                /// </summary>
                /// <param name="length">Length of block required</param>
                /// <returns>
                /// Offset of the allocated block, or 0 if there is not
                /// enough free space in the file
                /// </returns>
                public ulong Allocate(uint length)
                {
                    // Find a suitable block
                    LinkedListNode<FreeSpaceBlock> node = blocks.First;
                    LinkedListNode<FreeSpaceBlock> bestNode = null;
                    while (node != null)
                    {
                        if (node.Value.Length == length)
                        {
                            // We've found the best possible free block
                            bestNode = node;
                            break;
                        }
                        else if (node.Value.Length > length)
                        {
                            // Best fit algorithm - look for the smallest
                            // that fits
                            if (bestNode == null
                                || bestNode.Value.Length > node.Value.Length)
                            {
                                bestNode = node;
                            }
                        }

                        node = node.Next;
                    }

                    // Were we unable to find free space at all?
                    if (bestNode == null)
                    {
                        // TODO: May try to defragment
                        return 0;
                    }

                    // We found a block - let's cut it if necessary
                    if (length < bestNode.Value.Length)
                    {
                        SplitNode(bestNode, bestNode.Value.Offset + length);
                    }

                    // Remove the resulting block from the list
                    ulong allocatedBlockOffset = bestNode.Value.Offset;
                    blocks.Remove(bestNode);
                    freeTotal -= bestNode.Value.Length;

                    return allocatedBlockOffset;
                }

                /// <summary>
                /// Returns a block to the free space list
                /// </summary>
                /// <param name="offset">Offset of the block</param>
                /// <param name="length">Length of the block</param>
                public void Free(ulong offset, uint length) // TODO: Get rid of length?
                {
                    // Find where to add
                    LinkedListNode<FreeSpaceBlock> previousNode = null;
                    LinkedListNode<FreeSpaceBlock> nextNode = blocks.First;
                    while (nextNode != null && nextNode.Value.Offset < offset)
                    {
                        previousNode = nextNode;
                        nextNode = nextNode.Next;
                    }

                    // Check for overlap
                    if (previousNode != null
                            && previousNode.Value.Offset
                                + previousNode.Value.Length
                                > offset)
                    {
                        throw new ArgumentException(); // TODO: Fix exceptions
                    }
                    if (nextNode != null
                            && nextNode.Value.Offset < offset + length)
                    {
                        throw new ArgumentException(); // TODO: Fix exceptions
                    }

                    // We are going to free the block - adjust free space now,
                    // before we start merging and lose track of the original
                    // amount of space we're freeing
                    freeTotal += length;

                    // Merge with previous, if possible
                    if (previousNode != null
                            && offset
                                == previousNode.Value.Offset
                                    + previousNode.Value.Length)
                    {
                        offset = previousNode.Value.Offset;
                        length += previousNode.Value.Length;

                        blocks.Remove(previousNode);
                        previousNode = null;
                    }

                    // Merge with next, if possible
                    if (nextNode != null
                            && nextNode.Value.Offset == offset + length)
                    {
                        length += nextNode.Value.Length;

                        LinkedListNode<FreeSpaceBlock> newNextNode = nextNode.Next;
                        blocks.Remove(nextNode);
                        nextNode = newNextNode;
                    }

                    // Add the new block to the list
                    if (nextNode != null)
                    {
                        blocks.AddBefore(nextNode, new FreeSpaceBlock(offset, length));
                    }
                    else
                    {
                        blocks.AddLast(new FreeSpaceBlock(offset, length));
                    }
                }

                /// <summary>
                /// Excludes a data block from the free space list (used
                /// during file load)
                /// </summary>
                /// <param name="offset">Offset of the block</param>
                /// <param name="length">Length of the block</param>
                public void Exclude(ulong offset, uint length)
                {
                    // Find the free space block containing the one to exclude
                    // Find a suitable block
                    LinkedListNode<FreeSpaceBlock> node = blocks.First;
                    while (node != null
                            && node.Next != null
                            && node.Next.Value.Offset <= offset)
                    {
                        node = node.Next;
                    }

                    // Was there no free space block fully containing the
                    // block to exclude?
                    if (node == null
                            || node.Value.Offset > offset
                            || node.Value.Offset + node.Value.Length
                                < offset + length)
                    {
                        // Yes - throw exception only if there was free space
                        // partially containing the block to exclude. If
                        // there was no such free space at all, we're ok
                        // (we may be registering a descriptor that points
                        // to already allocated space, or we may be writing
                        // data at the end of the file).
                        if (node == null
                            || node.Value.Offset + node.Value.Length <= offset
                            || node.Value.Offset >= offset + length)
                        {
                            return;
                        }
                        throw new Exception(); // TODO: Fix exceptions
                    }

                    // We found a block

                    // Split if necessary
                    if (offset > node.Value.Offset)
                    {
                        SplitNode(node, offset);
                        node = node.Next;
                    }

                    // Cut if necessary
                    if (length < node.Value.Length)
                    {
                        SplitNode(node, offset + length);
                    }

                    // Remove the resulting node from the free block list
                    blocks.Remove(node);
                    freeTotal -= node.Value.Length;
                }

                /// <summary>
                /// 
                /// </summary>
                /// <param name="offset"></param>
                /// <param name="length"></param>
                public void GetFirst(out ulong offset, out uint length)
                {
                    if (blocks.First != null)
                    {
                        offset = blocks.First.Value.Offset;
                        length = blocks.First.Value.Length;
                    }
                    else
                    {
                        offset = 0;
                        length = 0;
                    }
                }

                /// <summary>
                /// 
                /// </summary>
                public ulong FreeSpaceTotal
                {
                    get
                    {
                        return freeTotal;
                    }
                }

                /// <summary>
                /// Splits a node in the free block list by cutting its block.
                /// The original node's block's length is decreased and a
                /// new node is inserted just after it for the remaining
                /// space.
                /// </summary>
                /// <param name="node">Node to split</param>
                /// <param name="offset">
                /// File offset to split at (must be inside the node's block)
                /// </param>
                private void SplitNode(LinkedListNode<FreeSpaceBlock> node,
                                       ulong offset)
                {
                    FreeSpaceBlock firstBlock;
                    FreeSpaceBlock secondBlock;
                    node.Value.Split(offset, out firstBlock, out secondBlock);

                    node.Value = firstBlock;
                    blocks.AddAfter(node, secondBlock);
                }

                LinkedList<FreeSpaceBlock> blocks;
                ulong freeTotal;
            }

            /// <summary>
            /// Data block manager - provides data block tracking and
            /// manipulation services
            /// </summary>
            private class DataBlockManager
            {
                /// <summary>
                /// 
                /// </summary>
                public DataBlockManager()
                {
                    descriptorsByOffset
                        = new SortedList<ulong, LinkedList<DataBlockDescriptor>>();
                    descriptorsByBlockOffset
                        = new SortedList<ulong, LinkedList<DataBlockDescriptor>>();
                }

                /// <summary>
                /// 
                /// </summary>
                /// <param name="descriptor"></param>
                public void RegisterDescriptor(DataBlockDescriptor descriptor)
                {
                    LinkedList<DataBlockDescriptor> sameBlockDescriptors;
                    bool otherDescriptorsForBlock
                        = descriptorsByBlockOffset.TryGetValue(
                            descriptor.BlockOffset,
                            out sameBlockDescriptors);

                    // Create new list of descriptors pointing to the same
                    // block, or verify that the already existing list is
                    // appropriate
                    if (!otherDescriptorsForBlock)
                    {
                        sameBlockDescriptors = new LinkedList<DataBlockDescriptor>();
                    }
                    else
                    {
                        if (descriptor.BlockLength
                                != sameBlockDescriptors.First.Value.BlockLength)
                        {
                            throw new Exception(); // TODO: Fix exceptions
                        }
                    }

                    // Add the descriptor to the list of same-offset descriptors
                    sameBlockDescriptors.AddLast(descriptor);

                    // In case we created a new list of same-offset descriptors,
                    // add it to the master list
                    if (!otherDescriptorsForBlock)
                    {
                        descriptorsByBlockOffset.Add(descriptor.BlockOffset, sameBlockDescriptors);
                    }
                }

                /// <summary>
                /// 
                /// </summary>
                /// <param name="descriptor"></param>
                /// <returns>
                /// True if this was the last descriptor at this offset
                /// </returns>
                public bool RemoveDescriptor(DataBlockDescriptor descriptor)
                {
                    LinkedList<DataBlockDescriptor> sameOffsetDescriptors;
                    bool descriptorsAtOffset
                        = descriptorsByBlockOffset.TryGetValue(
                            descriptor.BlockOffset,
                            out sameOffsetDescriptors);

                    if (!descriptorsAtOffset)
                    {
                        throw new Exception(); // TODO: Fix exceptions
                    }

                    if (!sameOffsetDescriptors.Remove(descriptor))
                    {
                        throw new Exception(); // TODO: Fix exceptions
                    }

                    if (sameOffsetDescriptors.Count > 0)
                    {
                        // We still have descriptors pointing to the same
                        // data block
                        return false;
                    }
                    else
                    {
                        descriptorsByBlockOffset.Remove(descriptor.BlockOffset);
                        return true;
                    }
                }

                /// <summary>
                /// 
                /// </summary>
                /// <param name="minOffset"></param>
                /// <param name="maxLength"></param>
                /// <returns></returns>
                public ulong GetLastDataBlockByMaxLength(
                    ulong minOffset,
                    uint maxLength)
                {
                    for (int i = descriptorsByBlockOffset.Count - 1; i >= 0; i--)
                    {
                        if (descriptorsByBlockOffset.Keys[i] >= minOffset
                            && descriptorsByBlockOffset.Values[i].First.Value.BlockLength <= maxLength)
                        {
                            return descriptorsByBlockOffset.Values[i].First.Value.BlockOffset;
                        }
                    }

                    return 0;
                }

                /// <summary>
                /// 
                /// </summary>
                /// <param name="offset"></param>
                /// <returns></returns>
                public uint GetDataBlockLength(ulong offset)
                {
                    return descriptorsByBlockOffset[offset].First.Value.BlockLength;
                }

                /// <summary>
                /// 
                /// </summary>
                /// <param name="offset"></param>
                /// <param name="newOffset"></param>
                /// <param name="controller">Slow operation controller</param>
                public void MoveBlock(
                    ulong offset,
                    ulong newOffset,
                    SlowOperationController controller)
                {
                    LinkedList<DataBlockDescriptor> sameOffsetDescriptors;
                    bool descriptorsAtOffset
                        = descriptorsByBlockOffset.TryGetValue(
                            offset,
                            out sameOffsetDescriptors);

                    if (!descriptorsAtOffset)
                    {
                        throw new Exception(); // TODO: Fix exceptions
                    }

                    DataBlockDescriptor moved = sameOffsetDescriptors.First.Value;
                    uint length = moved.MoveBlockTo(newOffset, controller);

                    // The above call to MoveBlockTo took care to remove
                    // the descriptor from this offset's list and file it
                    // properly with its new offset. We need to tell the
                    // other descriptors pointing to this block that they
                    // have a new data location. Calling AssignFrom will
                    // remove the descriptor from the list and refile it
                    // with its new location, so we only need to do this
                    // on the list's head.
                    while (sameOffsetDescriptors.First != null)
                    {
                        sameOffsetDescriptors.First.Value.AssignFrom(moved);
                    }

                    // Notify descriptors within the range that they have
                    // new offsets within the file
                    // TODO: Have a dedicated sorted structure and notify only descriptors that have actually moved!
                    foreach (LinkedList<DataBlockDescriptor> descriptorList in descriptorsByBlockOffset.Values)
                    {
                        foreach (DataBlockDescriptor descriptor in descriptorList)
                        {
                            descriptor.OnMove(offset, length, newOffset);
                        }
                    }
                }

                SortedList<ulong, LinkedList<DataBlockDescriptor>> descriptorsByOffset;
                SortedList<ulong, LinkedList<DataBlockDescriptor>> descriptorsByBlockOffset;
            }

            /// <summary>
            /// Initializes the Space Manger with the provided block of 
            /// free space
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="offset">offset of the free block</param>
            /// <param name="length">lengt of the free block</param>
            public SpaceManager(
                DataFileManager fileManager,
                ulong offset,
                uint length)
            {
                this.fileManager = fileManager;

                // Initialize free space
                freeBlocks = new FreeSpaceManager(offset, length);

                // Initialize the data block list
                dataBlocks = new DataBlockManager();
            }

            /// <summary>
            /// Allocates a block of the specified length from the free blocks
            /// </summary>
            /// <param name="length">Length of the desired block</param>
            /// <returns>
            /// Offset of the allocated block, or 0 if none of the free
            /// blocks is big enough
            /// </returns>
            public ulong AllocateDataBlock(uint length)
            {
                return freeBlocks.Allocate(length);
            }

            /// <summary>
            /// Frees a previously allocated block
            /// </summary>
            /// <param name="offset"></param>
            /// <param name="length"></param>
            private void FreeDataBlock(ulong offset, uint length)
            {
                freeBlocks.Free(offset, length);
            }

            /// <summary>
            /// Excludes a data block from the free space list (used
            /// during file load)
            /// </summary>
            /// <param name="offset">Offset of the block</param>
            /// <param name="length">Length of the block</param>
            public void ExcludeDataBlock(ulong offset, uint length)
            {
                freeBlocks.Exclude(offset, length);
            }

            /// <summary>
            /// Registers an allocated block's descriptor and tracks the
            /// space as allocated
            /// </summary>
            /// <param name="descriptor">Descriptor to register</param>
            public void RegisterDescriptor(DataBlockDescriptor descriptor)
            {
                // Store a reference to the descriptor
                dataBlocks.RegisterDescriptor(descriptor);

                // And track the space as not free
                if (descriptor.BlockLength > 0)
                {
                    freeBlocks.Exclude(
                        descriptor.BlockOffset,
                        descriptor.BlockLength);
                }
            }

            /// <summary>
            /// Un-registers an allocated block's descriptor
            /// </summary>
            /// <param name="descriptor">Descriptor to un-register</param>
            public void RemoveDescriptor(DataBlockDescriptor descriptor)
            {
                // Remove the descriptor from the list of descriptors
                if (dataBlocks.RemoveDescriptor(descriptor))
                {
                    freeBlocks.Free(descriptor.BlockOffset, descriptor.BlockLength);
                }
            }

            /// <summary>
            /// Defragments the free space to the file's end and truncates the
            /// file
            /// </summary>
            /// <param name="controller">Slow operation controller</param>
            public void Defragment(SlowOperationController controller)
            {
                ulong maxOffset
                    = (ulong)fileManager.stream.Length - freeBlocks.FreeSpaceTotal;

                ulong currentOffset;
                uint currentLength;

                // Do we need to defragment at all?
                if (freeBlocks.FreeSpaceTotal > 0)
                {
                    controller.StartProgress(maxOffset);
                    try
                    {
                        do
                        {
                            controller.ThrowIfCancelRequested();

                            // 1. Find the first free block
                            freeBlocks.GetFirst(out currentOffset, out currentLength);

                            controller.ReportProgress(currentOffset);

                            // 2. Find the last block of data that will fit into the free block
                            ulong blockOffset
                                = dataBlocks.GetLastDataBlockByMaxLength(
                                    currentOffset + currentLength,
                                    currentLength);

                            // Did we find a block that fits into the free block?
                            if (blockOffset != 0)
                            {
                                // Yes - 3. Move the block and repoint descriptors
                                dataBlocks.MoveBlock(
                                    blockOffset,
                                    currentOffset,
                                    controller);
                            }
                            else
                            {
                                // No
                                ulong nextBlockOffset = currentOffset + currentLength;

                                // Are there remaining data blocks at all?
                                if (nextBlockOffset == (ulong)fileManager.stream.Length)
                                {
                                    // No remaining data blocks - we're done
                                    break;
                                }

                                // There are more data blocks after the free block.
                                // Take the first data block immediately following
                                // the free block and move it somewhere else to free
                                // up space
                                uint nextBlockLength
                                    = dataBlocks.GetDataBlockLength(
                                        nextBlockOffset);

                                // Try moving the data block to another free block in
                                // the file
                                ulong allocatedOffset
                                    = freeBlocks.Allocate(nextBlockLength);
                                if (allocatedOffset == 0)
                                {
                                    // There is no free block big enough to hold the
                                    // data block - copy it at the end of the file.
                                    allocatedOffset = (ulong)fileManager.stream.Length;
                                }

                                // Move the block and repoint descriptors
                                dataBlocks.MoveBlock(
                                    nextBlockOffset,
                                    allocatedOffset,
                                    controller);
                            }
                        }
                        while (true);

                    }
                    finally
                    {
                        controller.FinishProgress();
                    }

                    // The free space should be in one free block at the end of the file
                    freeBlocks.GetFirst(out currentOffset, out currentLength);
                    if (currentOffset != 0)
                    {
                        // Tuncate it
                        fileManager.stream.SetLength((long)currentOffset);

                        // And remove this last standing free block from the list
                        freeBlocks.Exclude(currentOffset, currentLength);
                    }
                }
            }

            private DataFileManager fileManager;

            private FreeSpaceManager freeBlocks;
            private DataBlockManager dataBlocks;
        }

        /// <summary>
        /// Constructs a file manager on an already existing data file
        /// </summary>
        /// <param name="stream">Data file stream</param>
        public DataFileManager(Stream stream)
        {
            this.stream = stream;

            // We must initialize the space manager before reading ANY data
            spaceManager
                = new SpaceManager(
                    this,
                    0, // Everything after the header is free
                    (uint)stream.Length);

            if (stream.Length != 0)
            {
                // We're opening an already existing data file
                stream.Seek(0, SeekOrigin.Begin);
                fileHeader = new FileHeader(this, stream);

                // We just read the file header from the stream; remove it
                // from the free space list
                spaceManager.ExcludeDataBlock(0, (uint)stream.Position);
            }
            else
            {
                // We're creating a new data file
                fileHeader = new FileHeader(this);
                stream.Seek(0, SeekOrigin.Begin);
                fileHeader.WriteToStream(stream, false/*estimate*/);
            }

            // Get the custom data item array from the file (or an empty
            // array, if the file doesn't contain one)
            customData
                = fileHeader.CustomDataItemArrayDescriptor.BlockLength > 0
                    ? fileHeader.CustomDataItemArrayDescriptor.ReadData()
                    : new CustomDataItemArray(this);

            // Construct a new item list object on the file; it will take
            // care to read the list from the file
            items = new ItemList(this);
        }

        /// <summary>
        /// 
        /// </summary>
        ~DataFileManager()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// 
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (fileHeader != null)
                {
                    fileHeader.Dispose();
                    fileHeader = null;
                }

                if (customData != null)
                {
                    customData.Dispose();
                    customData = null;
                }

                if (items != null)
                {
                    items.Dispose();
                    items = null;
                }

                if (stream != null)
                {
                    stream.Close();
                    stream = null;
                }
            }
            // TODO: Exception?
        }

        /// <summary>
        /// Data file header
        /// </summary>
        public FileHeader FileHeader
        {
            get
            {
                return fileHeader;
            }
        }

        /// <summary>
        /// Item list
        /// </summary>
        public ItemList Items
        {
            get
            {
                return items;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public CustomDataItemArray CustomData
        {
            get
            {
                return customData;
            }
        }

        /// <summary>
        /// Allocates a data block with a specified length
        /// </summary>
        /// <param name="length">Length of the specified block</param>
        /// <returns>Offset of the newly allocated block</returns>
        private ulong AllocateDataBlock(uint length)
        {
            ulong offset = spaceManager.AllocateDataBlock(length);

            // Was there enough free space inside the file?
            if (offset > 0)
            {
                // Yes - use it
                return offset;
            }
            else
            {
                // No - write at the end of the file
                offset = (ulong)stream.Length;
            }

            return offset;
        }

        /// <summary>
        /// Excludes a data block from the free space list (used
        /// during file load)
        /// </summary>
        /// <param name="offset">Offset of the block</param>
        /// <param name="length">Length of the block</param>
        private void ExcludeDataBlock(ulong offset, uint length)
        {
            spaceManager.ExcludeDataBlock(offset, length);
        }

        /// <summary>
        /// Registers an allocated block's descriptor
        /// </summary>
        /// <param name="descriptor">Descriptor to register</param>
        private void RegisterDescriptor(DataBlockDescriptor descriptor)
        {
            spaceManager.RegisterDescriptor(descriptor);
        }

        /// <summary>
        /// Un-registers an allocated block's descriptor
        /// </summary>
        /// <param name="descriptor">Descriptor to un-register</param>
        private void RemoveDescriptor(DataBlockDescriptor descriptor)
        {
            spaceManager.RemoveDescriptor(descriptor);
        }

        /// <summary>
        /// 
        /// </summary>
        public void Defragment(SlowOperationController controller)
        {
            spaceManager.Defragment(controller);
        }

        // Data block length to signify that the item's contents have actually
        // been written to a separate external file instead of the main data file
        public const uint externalLength = uint.MaxValue;

        private SpaceManager spaceManager;

        private Stream stream;
        private FileHeader fileHeader;
        private ItemList items;
        private CustomDataItemArray customData;
    }
}
