using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Data Storage - stores and retrieves encrypted private items. The data
    /// storage knows nothing about the data itself - it only operates with
    /// streams (and the wrappers we have for them - Content and EncryptedContent)
    /// and item IDs. It also provides the means for saving/retrieving a small
    /// number of custom data items.
    /// </summary>
    public class DataStorage
    {
        /// <summary>
        /// Creates and opens a brand new storage
        /// </summary>
        /// <param name="dataPath">Path to the new main data file</param>
        /// <returns>DataStorage object representing the newly created storage</returns>
        public static DataStorage CreateStorage(string dataPath)
        {
            // Create a new empty data file
            FileStream newData
                = new FileStream(
                    dataPath,
                    FileMode.CreateNew,
                    FileAccess.ReadWrite,
                    FileShare.None);
            newData.Close();

            // Create the subfolder to hold external items
            Directory.CreateDirectory(ExternalFolderFromDataPath(dataPath));

            // Open the file
            return new DataStorage(dataPath);
        }

        /// <summary>
        /// Constructs DataStorage by given path to the main data file
        /// </summary>
        /// <param name="dataPath">Path to the main data file</param>
        public DataStorage(string dataPath)
        {
            this.dataPath = dataPath;
            externalItemFolder = ExternalFolderFromDataPath(dataPath);

            fileManager
                = new DataFileManager(
                    new FileStream(
                        dataPath,
                        FileMode.Open,
                        FileAccess.ReadWrite,
                        FileShare.None));
        }

        /// <summary>
        /// 
        /// </summary>
        ~DataStorage()
        {
            Close();
        }

        /// <summary>
        /// 
        /// </summary>
        public void Close()
        {
            if (fileManager != null)
            {
                fileManager.Dispose();
                fileManager = null;
            }
            dataPath = null;
        }

        /// <summary>
        /// Returns a list of IDs for all items in the storage
        /// </summary>
        /// <returns></returns>
        public ItemID[] GetItemIDs()
        {
            return fileManager.Items.ToArray();
        }

        /// <summary>
        /// Returns the custom data of a particular type
        /// </summary>
        /// <param name="type">Type of custom data to return</param>
        /// <returns>
        /// Custom data of the specified type. If the data is encrypted, the
        /// returned content is of type EncryptedContent.
        /// </returns>
        public Content GetCustomData(int type)
        {
            return fileManager.CustomData[type].Content;
        }

        /// <summary>
        /// Stores custom application data
        /// </summary>
        /// <param name="customData">Custom data to store</param>
        public void StoreCustomData(int type, Content content)
        {
            // Obtain a byte array containing the custom data
            Stream dataStream = content.DetachData();
            byte[] bytes = Utils.ReadStream(dataStream);
            dataStream.Close();

            // Write the byte array to the file
            DataFileManager.GenericDataBlockDescriptor contentDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            contentDescriptor.WriteData(bytes);

            // If the content is encrypted, store the initialization vector
            DataFileManager.GenericDataBlockDescriptor contentIVDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            if (content.GetType() == typeof(EncryptedContent))
            {
                contentIVDescriptor.WriteData(((EncryptedContent)content).IV.Bytes);
            }

            // Store the custom data item itself (containing descriptors to the
            // data we just saved)
            CustomDataItem item
                = new CustomDataItem(contentDescriptor, contentIVDescriptor);
            fileManager.CustomData.Store(type, item);
        }

        /// <summary>
        /// Retrieves the encrypted metadata for a specified private item.
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>Encrypted metadata for the item</returns>
        public EncryptedContent GetMetadata(ItemID id)
        {
            return fileManager.Items[id].Metadata;
        }

        /// <summary>
        /// Retreieves the encrypted contents of a specified private item.
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>Encrypted contents of the item</returns>
        public EncryptedContent GetContent(ItemID id)
        {
            EncryptedContent content = fileManager.Items[id].Content;

            // If the content is stored externally - read it
            if (content == null)
            {
                FileStream contentStream
                    = new FileStream(
                        ExternalFilePathFromID(id),
                        FileMode.Open, 
                        FileAccess.Read, 
                        FileShare.Read);
                content
                    = new EncryptedContent(
                        contentStream,
                        fileManager.Items[id].ContentIV);
            }

            return content;
        }

        /// <summary>
        /// Stores the encrypted metadata of an already existing private item.
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <param name="encryptedMetadata">Encrypted metadata for the item</param>
        /// <param name="controller">Slow operation controller</param>
        public void StoreMetadata(
            ItemID id,
            EncryptedContent encryptedMetadata,
            SlowOperationController controller)
        {
            controller.ThrowIfCancelRequested();

            if (!fileManager.Items.Contains(id))
            {
                // The item requested does not exist
                throw new Exception(); // TODO: Fix exceptions
            }

            // Obtain the item's content and content IV descriptors -
            // we will be preserving them
            DataFileManager.GenericDataBlockDescriptor contentBlock
                = fileManager.Items[id].ContentDescriptor;
            DataFileManager.GenericDataBlockDescriptor contentIVBlock
                = fileManager.Items[id].ContentIVDescriptor;

            // Store the item's metadata
            DataFileManager.GenericDataBlockDescriptor metadataBlock
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            Stream metadataStream = encryptedMetadata.DetachData();
            byte[] metadataBytes = Utils.ReadStream(metadataStream);
            metadataStream.Close();
            metadataBlock.WriteData(metadataBytes);

            // Store the item's metadata's IV
            DataFileManager.GenericDataBlockDescriptor metadataIVBlock
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            metadataIVBlock.WriteData(encryptedMetadata.IV.Bytes);

            // Add the item to the item list
            fileManager.Items.Store(
                id,
                new Item(
                    metadataBlock,
                    metadataIVBlock,
                    contentBlock,
                    contentIVBlock));
        }

        /// <summary>
        /// Stores the encrypted contents/metadata of a specified private item
        /// (if there is an item with the same ID already in the storage, it
        /// will be replaced.
        /// </summary>
        /// <param name="id">ID of the item to be stored</param>
        /// <param name="encryptedMetadata">Encrypted metadata for the item</param>
        /// <param name="encryptedContent">Encrypted contents of the item</param>
        /// <param name="isFile">
        /// Specifies whether the content is coming from a separate file (instead
        /// of being something the user typed in the program itself). We use
        /// this flag to decide whether the item should be stored as part of
        /// the main storage or externally.
        /// </param>
        /// <param name="controller">Slow operation controller</param>
        public void StoreItem(
            ItemID id,
            EncryptedContent encryptedMetadata,
            EncryptedContent encryptedContent,
            bool isFile,
            SlowOperationController controller)
        {
            controller.ThrowIfCancelRequested();

            // If we have content that is coming from an external file, we
            // have to assume it is of considerable size (i.e. more than
            // several hundred bytes); we will store this as a separate
            // encrypted file, outside of the main data file.
            if (isFile)
            {
                FileStream externalFile
                    = new FileStream(
                        ExternalFilePathFromID(id),
                        FileMode.Create,
                        FileAccess.Write,
                        FileShare.None);

                Stream contentStream = encryptedContent.DetachData();
                Utils.CopyWholeStreamToStream(
                    contentStream,
                    externalFile,
                    controller);
                contentStream.Close();

                externalFile.Close();
            }

            // Store the item's metadata
            DataFileManager.GenericDataBlockDescriptor metadataBlock
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            Stream metadataStream = encryptedMetadata.DetachData();
            byte[] metadataBytes = Utils.ReadStream(metadataStream);
            metadataStream.Close();
            metadataBlock.WriteData(metadataBytes);

            // Store the item's metadata's IV
            DataFileManager.GenericDataBlockDescriptor metadataIVBlock
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            metadataIVBlock.WriteData(encryptedMetadata.IV.Bytes);

            // Store the item's content
            DataFileManager.GenericDataBlockDescriptor contentBlock
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            byte[] contentBytes;
            if (!isFile)
            {
                Stream contentStream = encryptedContent.DetachData();
                contentBytes = Utils.ReadStream(contentStream);
                contentStream.Close();
            }
            else
            {
                // We already saved the contents externally
                contentBytes = null;
            }
            contentBlock.WriteData(contentBytes);

            // Store the item's content's IV
            DataFileManager.GenericDataBlockDescriptor contentIVBlock
                = new DataFileManager.GenericDataBlockDescriptor(fileManager);
            contentIVBlock.WriteData(encryptedContent.IV.Bytes);

            // Add the item to the item list
            fileManager.Items.Store(
                id,
                new Item(
                    metadataBlock,
                    metadataIVBlock,
                    contentBlock,
                    contentIVBlock));
        }

        /// <summary>
        /// Deletes a specified private item
        /// </summary>
        /// <param name="id">ID of the item to delete</param>
        public void DeleteItem(ItemID id)
        {
            // Let's check whether the item is external before we've deleted it
            bool external = fileManager.Items[id].External;

            // 1. Delete the item from the main data file
            fileManager.Items.Delete(id);

            // 2. If the item is external, delete the external data file as well
            if (external)
            {
                File.Delete(ExternalFilePathFromID(id));
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="controller">Slow operation controller</param>
        public void Defragment(SlowOperationController controller)
        {
            fileManager.Defragment(controller);
        }

        /// <summary>
        /// Calculates the name of the folder for storing external items
        /// </summary>
        /// <param name="dataPath">Path to the main data file</param>
        /// <returns>Path to the folder for storing external items</returns>
        private static string ExternalFolderFromDataPath(string dataPath)
        {
            string dataFolder = Path.GetDirectoryName(dataPath);
            if (dataFolder == null)
            {
                dataFolder = "";
            }

            return Path.Combine(
                    dataFolder,
                    Path.GetFileNameWithoutExtension(dataPath) + externalSubfolderSuffix);
        }

        /// <summary>
        /// Given an item's ID, returns the path to a file with a unique name
        /// to be used if the item's content were to be stored externally.
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>
        /// File path to use for storing the item's content externally
        /// </returns>
        private string ExternalFilePathFromID(ItemID id)
        {
            // We'll use the GUID from the item's ID to form the file name. By
            // using the "D" format specifier, the GUID will be formatted
            // without the brackets, but with dashes:
            // xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
            string externalFileName = id.ID.ToString("D") + externalExtension;

            return Path.Combine(externalItemFolder, externalFileName);
        }

        // Path to the main data file
        private string dataPath;
        private string externalItemFolder;

        private const string dataFileExtension = ".pds";
        private const string externalSubfolderSuffix = " Files";
        private const string externalExtension = ".pde";

        private DataFileManager fileManager;
    }
}
