using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Data File Header
    /// </summary>
    internal class FileHeader : IDisposable
    {
        private const int currentVersion = 1;

        /// <summary>
        /// Initializes a new file header
        /// </summary>
        /// <param name="fileManager">Data file manager</param>
        public FileHeader(DataFileManager fileManager)
        {
            customDataItemArrayDescriptor
                = new DataFileManager.CustomDataItemArrayDataBlockDescriptor(
                    fileManager);

            firstItemListElementDescriptor
                = new DataFileManager.ListElementDataBlockDescriptor(
                    fileManager);
        }

        /// <summary>
        /// Reads the file header from the main data file
        /// </summary>
        /// <param name="stream">Main data file</param>
        public FileHeader(DataFileManager fileManager, Stream stream)
        {
            // Read and check the version
            int version = Utils.ReadIntFromStream(stream);
            if (version != FileHeader.currentVersion)
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            // Offset of the custom data item array
            customDataItemArrayDescriptor
                = new DataFileManager.CustomDataItemArrayDataBlockDescriptor(
                    fileManager,
                    stream);

            // Offset of the first item list element
            firstItemListElementDescriptor
                = new DataFileManager.ListElementDataBlockDescriptor(
                    fileManager,
                    stream);
        }

        /// <summary>
        /// 
        /// </summary>
        ~FileHeader()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// 
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (customDataItemArrayDescriptor != null)
                {
                    customDataItemArrayDescriptor.Dispose();
                    customDataItemArrayDescriptor = null;
                }

                if (firstItemListElementDescriptor != null)
                {
                    firstItemListElementDescriptor.Dispose();
                    firstItemListElementDescriptor = null;
                }
            }
            // TODO: Exception?
        }

        /// <summary>
        /// Writes the file header to a stream
        /// </summary>
        /// <param name="stream">Stream to write the file header to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public uint WriteToStream(Stream stream, bool estimate)
        {
            uint bytesWritten = 0;

            // Version
            bytesWritten += Utils.WriteIntToStream(currentVersion, stream, estimate);

            // Custom data item array descriptor
            bytesWritten
                += customDataItemArrayDescriptor.WriteToStream(
                    stream,
                    estimate);

            // First item list element descriptor
            bytesWritten
                += firstItemListElementDescriptor.WriteToStream(
                    stream,
                    estimate);

            // Keep file in consistent state
            if (!estimate)
            {
                stream.Flush();
            }

            return bytesWritten;
        }

        /// <summary>
        /// Descriptor of the array of custom data items
        /// </summary>
        public DataFileManager.CustomDataItemArrayDataBlockDescriptor CustomDataItemArrayDescriptor
        {
            get
            {
                return customDataItemArrayDescriptor;
            }
        }

        /// <summary>
        /// Descriptor of the first item list element's data block
        /// </summary>
        public DataFileManager.ListElementDataBlockDescriptor FirstItemListElementDescriptor
        {
            get
            {
                return firstItemListElementDescriptor;
            }
        }

        private DataFileManager.CustomDataItemArrayDataBlockDescriptor
            customDataItemArrayDescriptor;
        private DataFileManager.ListElementDataBlockDescriptor
            firstItemListElementDescriptor;
    }
}
