using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Represents a single item stored in the storage
    /// </summary>
    internal class Item : IDisposable
    {
        /// <summary>
        /// Constructs an item object
        /// </summary>
        /// <param name="metadata">
        /// Descriptor for the data block containing the item's metadata
        /// </param>
        /// <param name="metadataIV">
        /// Descriptor for the data block containing the item's metadata's IV
        /// </param>
        /// <param name="content">
        /// Descriptor for the data block containing the item's contents
        /// </param>
        /// <param name="contentIV">
        /// Descriptor for the data block containing the item's contents' IV
        /// </param>
        public Item(DataFileManager.GenericDataBlockDescriptor metadata,
                    DataFileManager.GenericDataBlockDescriptor metadataIV,
                    DataFileManager.GenericDataBlockDescriptor content,
                    DataFileManager.GenericDataBlockDescriptor contentIV)
        {
            metadataDescriptor = metadata;
            metadataIVDescriptor = metadataIV;
            contentDescriptor = content;
            contentIVDescriptor = contentIV;
        }

        /// <summary>
        /// Reads the item record from a stream
        /// </summary>
        /// <param name="fileManager">Data file manager</param>
        /// <param name="stream">Stream to read from</param>
        public Item(DataFileManager fileManager, Stream stream)
        {
            metadataDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(
                    fileManager,
                    stream);
            metadataIVDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(
                    fileManager,
                    stream);
            contentDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(
                    fileManager,
                    stream);
            contentIVDescriptor
                = new DataFileManager.GenericDataBlockDescriptor(
                    fileManager,
                    stream);
        }

        /// <summary>
        /// Destroys the object
        /// </summary>
        ~Item()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// Disposes the resources held by the object (inherited from IDisposable)
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                metadataDescriptor.Dispose();
                metadataIVDescriptor.Dispose();
                contentDescriptor.Dispose();
                contentIVDescriptor.Dispose();
            }
#if DEBUG
            else
            {
                // Somebody forgot to call Dispose explicitly!!!
                throw new Exception(); // TODO: Fix exceptions
            }
#endif
        }

        /// <summary>
        /// Writes the item record to a stream
        /// </summary>
        /// <param name="stream">Stream to write to</param>
        /// <param name="estimate">
        /// Estimate space required only; do not actually write
        /// </param>
        /// <returns>Bytes written (or required, if estimating)</returns>
        public uint WriteToStream(Stream stream, bool estimate)
        {
            uint bytesWritten = 0;

            bytesWritten += metadataDescriptor.WriteToStream(stream, estimate);
            bytesWritten += metadataIVDescriptor.WriteToStream(stream, estimate);
            bytesWritten += contentDescriptor.WriteToStream(stream, estimate);
            bytesWritten += contentIVDescriptor.WriteToStream(stream, estimate);

            // Keep file in consistent state
            if (!estimate)
            {
                stream.Flush();
            }

            return bytesWritten;
        }

        /// <summary>
        /// Gets the item's metadata
        /// </summary>
        public EncryptedContent Metadata
        {
            get
            {
                return
                    new EncryptedContent(
                        new MemoryStream(metadataDescriptor.ReadData()),
                        new IV(metadataIVDescriptor.ReadData()));
            }
        }

        /// <summary>
        /// Gets the item's content; returns null if the content is saved
        /// externally to the main data file.
        /// </summary>
        public EncryptedContent Content
        {
            get
            {
                if (!External)
                {
                    // The item's content is in the main data file
                    return
                        new EncryptedContent(
                            new MemoryStream(contentDescriptor.ReadData()),
                            new IV(contentIVDescriptor.ReadData()));
                }
                else
                {
                    // The item's content is saved externally
                    return null;
                }
            }
        }

        /// <summary>
        /// Gets the IV for the item's content (needed for externally saved
        /// content)
        /// </summary>
        public IV ContentIV
        {
            get
            {
                return new IV(contentIVDescriptor.ReadData());
            }
        }

        /// <summary>
        /// Is the item external?
        /// </summary>
        public bool External
        {
            get
            {
                return contentDescriptor.BlockLength == DataFileManager.externalLength;
            }
        }

        /// <summary>
        /// Returns the descriptor to the item's content
        /// </summary>
        public DataFileManager.GenericDataBlockDescriptor ContentDescriptor
        {
            get
            {
                return contentDescriptor;
            }
        }

        /// <summary>
        /// Returns the descriptor to the item's content's IV
        /// </summary>
        public DataFileManager.GenericDataBlockDescriptor ContentIVDescriptor
        {
            get
            {
                return contentIVDescriptor;
            }
        }

        private DataFileManager.GenericDataBlockDescriptor metadataDescriptor;
        private DataFileManager.GenericDataBlockDescriptor metadataIVDescriptor;
        private DataFileManager.GenericDataBlockDescriptor contentDescriptor;
        private DataFileManager.GenericDataBlockDescriptor contentIVDescriptor;
    }
}
