using System;
using System.Collections;
using System.Text;
using System.IO;

using PrivateDataCommon;

namespace PrivateDataStorage
{
    /// <summary>
    /// Represents a list of all the items stored
    /// </summary>
    internal class ItemList : IDisposable
    {
        /// <summary>
        /// Represents a single list element
        /// </summary>
        public class ListElement : IDisposable
        {
            /// <summary>
            /// 
            /// </summary>
            /// <param name="nextElementDescriptor"></param>
            /// <param name="id"></param>
            /// <param name="item"></param>
            public ListElement(
                DataFileManager.ListElementDataBlockDescriptor nextElementDescriptor,
                ItemID id,
                Item item)
            {
                this.nextElementDescriptor = nextElementDescriptor;
                this.id = id;
                this.item = item;
            }

            /// <summary>
            /// Reads an item list element from the main data file's stream
            /// </summary>
            /// <param name="fileManager">Data file manager</param>
            /// <param name="stream">Main data file</param>
            public ListElement(
                DataFileManager fileManager,
                Stream stream)
            {
                // Read the descriptor of the next list item
                nextElementDescriptor
                    = new DataFileManager.ListElementDataBlockDescriptor(
                        fileManager,
                        stream);

                // Read the item's ID
                id = new ItemID(stream);

                // Read the item itself
                item = new Item(fileManager, stream);
            }

            /// <summary>
            /// 
            /// </summary>
            ~ListElement()
            {
                Dispose(false/*disposing*/);
            }

            /// <summary>
            /// 
            /// </summary>
            public void Dispose()
            {
                Dispose(true/*disposing*/);
                GC.SuppressFinalize(this);
            }

            /// <summary>
            /// Disposes the resources held by the object
            /// </summary>
            /// <param name="disposing">
            /// True if this method is called directly from user code. False if
            /// Dispose has been called by the runtime through the class' finalizer
            /// (in this case, we can't dereference / try to dispose resources
            /// held by the object as they may have already been destroyed).
            /// </param>
            protected virtual void Dispose(bool disposing)
            {
                if (disposing)
                {
                    if (nextElementDescriptor != null)
                    {
                        nextElementDescriptor.Dispose();
                        nextElementDescriptor = null;
                    }

                    if (item != null)
                    {
                        item.Dispose();
                        item = null;
                    }
                }
                // TODO: Exception?
            }

            /// <summary>
            /// Writes the item list element to a stream
            /// </summary>
            /// <param name="stream">Stream to write to</param>
            /// <param name="estimate">
            /// Estimate space required only; do not actually write
            /// </param>
            /// <returns>Bytes written (or required, if estimating)</returns>
            public uint WriteToStream(Stream stream, bool estimate)
            {
                uint bytesWritten = 0;

                bytesWritten
                    += nextElementDescriptor.WriteToStream(stream, estimate);
                bytesWritten += id.WriteToStream(stream, estimate);
                bytesWritten += item.WriteToStream(stream, estimate);

                // Keep file in consistent state
                if (!estimate)
                {
                    stream.Flush();
                }

                return bytesWritten;
            }

            /// <summary>
            /// Descriptor for the next list element
            /// </summary>
            public DataFileManager.ListElementDataBlockDescriptor NextElementDescriptor
            {
                get
                {
                    return nextElementDescriptor;
                }
            }

            /// <summary>
            /// ID of the item described by the list element
            /// </summary>
            public ItemID ID
            {
                get
                {
                    return id;
                }
            }

            /// <summary>
            /// Item described by the list element
            /// </summary>
            public Item Item
            {
                get
                {
                    return item;
                }
            }

            /// <summary>
            /// Previous list element in the file
            /// </summary>
            public ListElement Previous
            {
                get
                {
                    return previous;
                }

                set
                {
                    previous = value;
                }
            }

            /// <summary>
            /// Next list element in the file
            /// </summary>
            public ListElement Next
            {
                get
                {
                    return next;
                }

                set
                {
                    next = value;
                }
            }

            private ListElement previous;
            private ListElement next;

            private DataFileManager.ListElementDataBlockDescriptor nextElementDescriptor;
            private ItemID id;
            private Item item;
        }

        /// <summary>
        /// Constructs an item list on the main data file
        /// </summary>
        /// <param name="fileManager">Data file manager</param>
        public ItemList(DataFileManager fileManager)
        {
            this.fileManager = fileManager;
            items = new Hashtable();

            // If the file contains an item list, read it in
            DataFileManager.ListElementDataBlockDescriptor nextListElementDescriptor
                = fileManager.FileHeader.FirstItemListElementDescriptor;
            ListElement previousListElement = null;
            while (nextListElementDescriptor.BlockLength != 0)
            {
                ListElement listElement = nextListElementDescriptor.ReadData();

                // Establish a link with the previous
                if (previousListElement != null)
                {
                    previousListElement.Next = listElement;
                }
                else
                {
                    firstListElement = listElement;
                }
                listElement.Previous = previousListElement;
                previousListElement = listElement;

                // And add the list element into the hash table
                items.Add(listElement.ID, listElement);

                // Finally, see where we have to read the next list element from
                nextListElementDescriptor = listElement.NextElementDescriptor;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        ~ItemList()
        {
            Dispose(false/*disposing*/);
        }

        /// <summary>
        /// 
        /// </summary>
        public void Dispose()
        {
            Dispose(true/*disposing*/);
        }

        /// <summary>
        /// Disposes the resources held by the object
        /// </summary>
        /// <param name="disposing">
        /// True if this method is called directly from user code. False if
        /// Dispose has been called by the runtime through the class' finalizer
        /// (in this case, we can't dereference / try to dispose resources
        /// held by the object as they may have already been destroyed).
        /// </param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (items != null)
                {
                    foreach (ListElement listElement in items.Values)
                    {
                        listElement.Dispose();
                    }
                }
            }
            // TODO: Exception?
        }

        /// <summary>
        /// Adds/modifies an item
        /// </summary>
        /// <param name="id">ID of the item to store</param>
        /// <param name="item">Item to store</param>
        public void Store(ItemID id, Item item)
        {
            // Create a new list element
            ListElement listElement
                = new ListElement(
                    new DataFileManager.ListElementDataBlockDescriptor(fileManager),
                    id,
                    item);

            // Write it out
            DataFileManager.ListElementDataBlockDescriptor listElementDescriptor
                = new DataFileManager.ListElementDataBlockDescriptor(fileManager);
            listElementDescriptor.WriteData(listElement);

            // Prepare it for inclusion in the list by setting its
            // "previous" and "next" pointers appropriately

            ListElement oldListElement
                = items.Contains(id)
                    ? (ListElement)items[id]
                    : null;

            if (oldListElement != null)
            {
                listElement.Previous = oldListElement.Previous;
                listElement.Next = oldListElement.Next;
            }
            else
            {
                // Insert new elements at the beginning of the list
                listElement.Previous = null;
                listElement.Next = firstListElement;
            }

            // And include it in the list itself; again, we're adding
            // new elements at the front of the list
            DataFileManager.ListElementDataBlockDescriptor descriptorToRepoint
                = listElement.Previous != null
                    ? listElement.Previous.NextElementDescriptor
                    : fileManager.FileHeader.FirstItemListElementDescriptor;

            // Replace the old element with the new one
            // (or insert the new one, if it's new to the list)
            items[id] = listElement;
            
            if (listElement.Previous == null)
            {
                firstListElement = listElement;
            }
            else
            {
                listElement.Previous.Next = listElement;
            }

            if (listElement.Next != null)
            {
                listElement.Next.Previous = listElement;
            }

            // And do the same in the file by repointing the descriptors
            listElement.NextElementDescriptor.AssignFrom(
                oldListElement != null
                    ? oldListElement.NextElementDescriptor
                    : descriptorToRepoint);
            descriptorToRepoint.AssignFrom(listElementDescriptor);

            // Finally, dispose the list element we're replacing
            if (oldListElement != null)
            {
                oldListElement.Dispose();
                oldListElement = null;
            }

            // We have already assigned this descriptor away, so we no
            // longer need it
            listElementDescriptor.Dispose();
        }

        /// <summary>
        /// Deletes an item
        /// </summary>
        /// <param name="id">ID of the item to be deleted</param>
        public void Delete(ItemID id)
        {
            if (!items.Contains(id))
            {
                throw new Exception(); // TODO: Fix exceptions
            }

            ListElement listElement = ((ListElement)items[id]);

            // Free the blocks for the item's data if not needed
            listElement.Item.Dispose();

            // Exclude the item from the list of items
            {
                DataFileManager.ListElementDataBlockDescriptor descriptorToRepoint
                    = listElement.Previous != null
                        ? listElement.Previous.NextElementDescriptor
                        : fileManager.FileHeader.FirstItemListElementDescriptor;

                // Exclude it from list in file
                descriptorToRepoint.AssignFrom(listElement.NextElementDescriptor);

                // Exclude it from list in memory
                items.Remove(id);
                
                if (listElement.Previous != null)
                {
                    listElement.Previous.Next = listElement.Next;
                }
                else
                {
                    firstListElement = listElement.Next;
                }

                if (listElement.Next != null)
                {
                    listElement.Next.Previous = listElement.Previous;
                }
            }

            listElement.Dispose();
        }

        /// <summary>
        /// Iterator by item ID
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>Item, if found; otherwise null</returns>
        public Item this[ItemID id]
        {
            get
            {
                return items.Contains(id) ? ((ListElement)items[id]).Item : null;
            }
        }

        /// <summary>
        /// Creates a new enumerator by ItemID
        /// </summary>
        /// <returns>Enumerator by ItemID</returns>
        public IEnumerator GetEnumerator()
        {
            return items.Keys.GetEnumerator();
        }

        /// <summary>
        /// Converts the item list to an array of item IDs
        /// </summary>
        /// <returns>Array of item IDs</returns>
        public ItemID[] ToArray()
        {
            ItemID[] ids = new ItemID[items.Count];

            int i = 0;
            foreach (ItemID id in items.Keys)
            {
                ids[i++] = id;
            }

            return ids;
        }

        /// <summary>
        /// Number of items
        /// </summary>
        public int Count
        {
            get
            {
                return items.Count;
            }
        }

        /// <summary>
        /// Returns whether an item with a specified ID already exists
        /// </summary>
        /// <param name="id">ID of the item</param>
        /// <returns>true if the item already exists; false if not</returns>
        public bool Contains(ItemID id)
        {
            return items.ContainsKey(id);
        }

        // Data file manager
        private DataFileManager fileManager;

        private Hashtable items;
        private ListElement firstListElement;
    }
}
