using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

using PrivateDataCommon;
using PrivateDataCommon.Exceptions;

namespace PrivateData
{
    public abstract class AsyncOperation
    {
        public AsyncOperation()
        {
            controller = new SlowOperationController();
            controller.ProgressStarted
                += new EventHandler<ProgressStartedArgs>(
                        controller_ProgressStarted);
            controller.ProgressReported
                += new EventHandler<ProgressReportedArgs>(
                        controller_ProgressReported);
            controller.ProgressFinished
                += new EventHandler(controller_ProgressFinished);
        }

        public abstract string Message
        {
            get;
        }

        public abstract bool Cancellable
        {
            get;
        }

        public void BeginExecute()
        {
            Thread thread = new Thread(new ThreadStart(ThreadMain));
            thread.Start();
        }

        protected abstract void Execute(SlowOperationController controller);

        private void ThreadMain()
        {
            try
            {
                lock (this)
                {
                    status = Status.InProgress;
                }

                Execute(controller);

                lock (this)
                {
                    status = Status.Succeeded;
                    OnSucceeded(EventArgs.Empty);
                }
            }
            catch (OperationCancelledException)
            {
                lock (this)
                {
                    status = Status.Cancelled;
                    OnCancelled(EventArgs.Empty);
                }
            }
            catch
            {
                lock (this)
                {
                    status = Status.Failed;
                    OnFailed(EventArgs.Empty);
                }
            }
        }

        public bool OperationSucceeded
        {
            get
            {
                lock (this)
                {
                    return status == Status.Succeeded;
                }
            }
        }

        public bool OperationCancelled
        {
            get
            {
                lock (this)
                {
                    return status == Status.Cancelled;
                }
            }
        }

        public bool OperationFailed
        {
            get
            {
                lock (this)
                {
                    return status == Status.Failed;
                }
            }
        }

        public bool OperationDone
        {
            get
            {
                lock (this)
                {
                    return status != Status.None && status != Status.InProgress;
                }
            }
        }

        public bool Wait()
        {
            while (!OperationDone)
            {
                // Unfortunately we don't have Monitor.Wait and Pulse in
                // the Compact Framework, so we'll just poll the flag
                // periodically until it's set.
                // Monitor.Wait(1000);
                Thread.Sleep(200);
            }

            return OperationSucceeded;
        }

        public void Cancel()
        {
            controller.CancelRequested = true;
        }

        public bool CancelAndWait()
        {
            Cancel();

            // This must be outside a lock, otherwise we will deadlock -
            // Wait needs to release the lock on the object so that the worker
            // thread can set the status when it's done. Normally this would
            // have been achieved through Monitor.Wait as it releases the lock,
            // but since we don't have that in the Compact Framework, Wait does
            // it itself. Therefore, we shouldn't have a lock around it.
            Wait();

            return !OperationSucceeded;
        }

        private void controller_ProgressStarted(
            object sender,
            ProgressStartedArgs e)
        {
            OnProgressStarted(e);
        }

        private void controller_ProgressReported(
            object sender,
            ProgressReportedArgs e)
        {
            OnProgressReported(e);
        }

        private void controller_ProgressFinished(
            object sender,
            EventArgs e)
        {
            OnProgressFinished(e);
        }

        protected void OnSucceeded(EventArgs e)
        {
            // TODO: Is this method thread-safe?
            if (Succeeded != null)
            {
                // TODO: Raise asynchronously and/or on the correct thread!
                Succeeded(this, e);
            }
        }

        protected void OnCancelled(EventArgs e)
        {
            // TODO: Is this method thread-safe?
            if (Cancelled != null)
            {
                // TODO: Raise asynchronously and/or on the correct thread!
                Cancelled(this, e);
            }
        }

        protected void OnFailed(EventArgs e)
        {
            // TODO: Is this method thread-safe?
            if (Failed != null)
            {
                // TODO: Raise asynchronously and/or on the correct thread!
                Failed(this, e);
            }
        }

        protected void OnProgressStarted(ProgressStartedArgs e)
        {
            // TODO: Is this method thread-safe?
            if (ProgressStarted != null)
            {
                // TODO: Raise asynchronously and/or on the correct thread!
                ProgressStarted(this, e);
            }
        }

        protected void OnProgressReported(ProgressReportedArgs e)
        {
            // TODO: Is this method thread-safe?
            if (ProgressReported != null)
            {
                // TODO: Raise asynchronously and/or on the correct thread!
                ProgressReported(this, e);
            }
        }

        protected void OnProgressFinished(EventArgs e)
        {
            // TODO: Is this method thread-safe?
            if (ProgressFinished != null)
            {
                // TODO: Raise asynchronously and/or on the correct thread!
                ProgressFinished(this, e);
            }
        }

        private enum Status
        {
            None,
            InProgress,
            Succeeded,
            Cancelled,
            Failed
        };

        private Status status;

        protected SlowOperationController controller;

        public event EventHandler Succeeded;
        public event EventHandler Cancelled;
        public event EventHandler Failed;

        public event EventHandler<ProgressStartedArgs> ProgressStarted;
        public event EventHandler<ProgressReportedArgs> ProgressReported;
        public event EventHandler ProgressFinished;
    }
}
