using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

using PrivateDataCommon;

namespace PrivateData
{
    internal class AsyncOperationProgressTracker
    {
        public AsyncOperationProgressTracker(
            AsyncOperation operation,
            ProgressBar progressBar)
        {
            this.progressBar = progressBar;

            progressLevels = new Stack<ProgressLevel>();
            progressMax = 1;
            progressCurrent = 0;

            operation.ProgressStarted
                += new EventHandler<ProgressStartedArgs>(operation_ProgressStarted);
            operation.ProgressReported
                += new EventHandler<ProgressReportedArgs>(operation_ProgressReported);
            operation.ProgressFinished
                += new EventHandler(operation_ProgressFinished);
        }

        private void operation_ProgressStarted(object sender, ProgressStartedArgs e)
        {
            progressBar.BeginInvoke(
                new EventHandler<ProgressStartedArgs>(ProgressStarted),
                sender,
                e);
        }

        private void operation_ProgressReported(object sender, ProgressReportedArgs e)
        {
            progressBar.BeginInvoke(
                new EventHandler<ProgressReportedArgs>(ProgressReported),
                sender,
                e);
        }

        private void operation_ProgressFinished(object sender, EventArgs e)
        {
            progressBar.BeginInvoke(
                new EventHandler(ProgressFinished),
                sender,
                e);
        }

        private void ProgressStarted(object sender, ProgressStartedArgs e)
        {
            // Push a new progress precision level on top of the stack
            progressLevels.Push(
                new ProgressLevel(
                    progressCurrent * e.ProgressMax,
                    e.ProgressMax));

            // Increase the precision of progress tracking
            progressCurrent *= e.ProgressMax;
            progressMax *= e.ProgressMax;
        }

        private void ProgressReported(object sender, ProgressReportedArgs e)
        {
            // Update the current progress position
            progressCurrent = progressLevels.Peek().Position + e.Progress;

            // And update the status bar, converting the progress position
            // to the real progress bar's precision
            UpdateProgress();
        }

        private void ProgressFinished(object sender, EventArgs e)
        {
            // Pop the finished progress precision level from the stack
            ProgressLevel level = progressLevels.Pop();

            // Return the precision of progress tracking to its previous level.
            // When doing this, we'll also move further a notch since we just
            // finished a whole new precision level, which should have been
            // equivalent to a notch of the level above it
            progressCurrent = level.Position / level.Max + 1;
            progressMax /= level.Max;

            // And update the status bar, converting the progress position
            // to the real progress bar's precision
            UpdateProgress();
        }

        private void UpdateProgress()
        {
#if DEBUG
            if (progressCurrent > progressMax)
            {
                throw new Exception(); // TODO: Fix exceptions
            }
#endif

            // Limit current progress to progressMax
            ulong current = Math.Min(progressCurrent, progressMax);

            // Update the progress bar, converting the progress position to the
            // real progress bar's precision
            progressBar.Value = (int)(current * (uint)progressBar.Maximum / progressMax);
        }

        private class ProgressLevel
        {
            public ProgressLevel(ulong position, ulong max)
            {
                this.position = position;
                this.max = max;
            }

            public ulong Position
            {
                get
                {
                    return position;
                }

                set
                {
                    position = value;
                }
            }

            public ulong Max
            {
                get
                {
                    return max;
                }
            }

            private ulong position;
            private ulong max;
        }

        private ProgressBar progressBar;

        private Stack<ProgressLevel> progressLevels;
        private ulong progressMax;
        private ulong progressCurrent;
    }
}
