using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.IO;

using PrivateDataCommon;
using PrivateDataCore;
using PrivateDataCore.ItemTypes;

namespace PrivateData
{
    public partial class EditItemDialog : Form
    {
        static EditItemDialog()
        {
            itemControlConstructors = new Dictionary<ItemType, ConstructorInfo>();

            itemControlConstructors.Add(
                ItemType.Note,
                GetItemControlConstructorInfo(typeof(ItemControls.NoteControl)));
            itemControlConstructors.Add(
                ItemType.GenericPassword,
                GetItemControlConstructorInfo(typeof(ItemControls.GenericPasswordControl)));
            itemControlConstructors.Add(
                ItemType.WebAccount,
                GetItemControlConstructorInfo(typeof(ItemControls.WebAccountControl)));
            itemControlConstructors.Add(
                ItemType.Email,
                GetItemControlConstructorInfo(typeof(ItemControls.EmailControl)));
            itemControlConstructors.Add(
                ItemType.BankAccount,
                GetItemControlConstructorInfo(typeof(ItemControls.BankAccountControl)));
            itemControlConstructors.Add(
                ItemType.CreditCard,
                GetItemControlConstructorInfo(typeof(ItemControls.CreditCardControl)));
            itemControlConstructors.Add(
                ItemType.CellPhone,
                GetItemControlConstructorInfo(typeof(ItemControls.CellPhoneControl)));
            itemControlConstructors.Add(
                ItemType.Picture,
                GetItemControlConstructorInfo(typeof(ItemControls.PictureControl)));
            itemControlConstructors.Add(
                ItemType.File,
                GetItemControlConstructorInfo(typeof(ItemControls.FileControl)));
        }

        private static ConstructorInfo GetItemControlConstructorInfo(Type type)
        {
            // Get the type's public instance constructors (we need to call
            // GetConstructors instead of GetConstructor so that we can use the
            // "is" operator below as we don't know the precise type of the
            // parameter
            ConstructorInfo[] constructors
                = type.GetConstructors(
                    BindingFlags.Public | BindingFlags.Instance);

            // Iterate through the constructors and find the one we want
            foreach (ConstructorInfo constructor in constructors)
            {
                ParameterInfo[] parameters = constructor.GetParameters();

                if (parameters.Length == 1
                    && parameters[0].ParameterType.BaseType == typeof(Item))
                {
                    return constructor;
                }
            }

            // Internal error: we couldn't find the constructor
            throw new Exception(); // TODO: Fix exceptions
        }

        public EditItemDialog(DataManager dataManager, Item item, ItemID folder)
        {
            this.dataManager = dataManager;
            this.item = item;
            this.folder = folder;

            InitializeComponent();
            CreateEditItemControl();
            if (item is IFileItem)
            {
                exportButton.Visible = !((IFileItem)item).IsImported; // TOOD: anything else?
                okButton.Enabled = ((IFileItem)item).IsImported;
            }
        }

        public void CreateEditItemControl()
        {
            UserControl control
                = (UserControl)itemControlConstructors[item.Type].Invoke(
                    new object[] { item });

            SuspendLayout();
            control.Dock = DockStyle.Fill;
            editItemPanel.Controls.Add(control);
            ResumeLayout();

            editItemControl = (IEditItemControl)control;
        }

        private void EditItemDialog_Resize(object sender, EventArgs e)
        {
            AdjustFormPanel();
        }

        private void inputPanel_EnabledChanged(object sender, EventArgs e)
        {
            AdjustFormPanel();
        }

        private void AdjustFormPanel()
        {
            formPanel.SuspendLayout();

            if (inputPanel.Enabled)
            {
                Rectangle visibleDesktop
                    = RectangleToClient(inputPanel.VisibleDesktop);

                formPanel.Dock = DockStyle.None;
                formPanel.Location = visibleDesktop.Location;
                formPanel.Size = visibleDesktop.Size;
            }
            else
            {
                formPanel.Dock = DockStyle.Fill;
            }

            formPanel.ResumeLayout();
        }

        private void exportButton_Click(object sender, EventArgs e)
        {
            exportButton.Enabled = false;
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.FileName = item.FileName;
            string extension = Path.GetExtension(item.FileName);
            saveFileDialog.Filter = "*" + extension + "|*" + extension;

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                AsyncOperations.ExportItemOperation exportItem
                    = new AsyncOperations.ExportItemOperation(
                        item as IFileItem,
                        saveFileDialog.FileName);

                new AsyncOperationUI(
                    exportItem,
                    this,
                    new EventHandler(exportItem_Succeeded),
                    new EventHandler(exportItem_Cancelled),
                    new EventHandler(exportItem_Failed));

                exportItem.BeginExecute();
            }
            else
            {
                exportButton.Enabled = true;
            }
        }

        private void okButton_Click(object sender, EventArgs e)
        {
            if (editItemControl.Save())
            {
                // TODO: File into proper folder
                AsyncOperations.StoreItemOperation storeItem =
                    new AsyncOperations.StoreItemOperation(
                        dataManager,
                        item,
                        folder);

                new AsyncOperationUI(
                    storeItem,
                    this,
                    new EventHandler(storeItem_Succeeded),
                    new EventHandler(storeItem_Cancelled),
                    new EventHandler(storeItem_Failed));

                storeItem.BeginExecute();
            }
            else
            {
                // We don't need to save - just close the dialog
                DialogResult = DialogResult.OK;
                Close();
            }
        }

        private void storeItem_Succeeded(object sender, EventArgs e)
        {
            DialogResult = DialogResult.OK;
            Close();
        }

        private void storeItem_Cancelled(object sender, EventArgs e)
        {
            DialogResult = DialogResult.Cancel;
            Close();
        }

        private void storeItem_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Failed to store item");
            
            DialogResult = DialogResult.Abort;
            Close();
        }

        private void exportItem_Succeeded(object sender, EventArgs e)
        {
            MessageBox.Show(
                "Export succeeded",
                Properties.Resources.MessageBoxCaptionSuccess,
                MessageBoxButtons.OK,
                MessageBoxIcon.Asterisk,
                MessageBoxDefaultButton.Button1);
            Close();
        }

        private void exportItem_Cancelled(object sender, EventArgs e)
        {
            MessageBox.Show(
                "Export cancelled",
                Properties.Resources.MessageBoxCaptionCancelled,
                MessageBoxButtons.OK,
                MessageBoxIcon.Exclamation,
                MessageBoxDefaultButton.Button1);
            Close();
        }

        private void exportItem_Failed(object sender, EventArgs e)
        {
            MessageBox.Show(
                "Export failed",
                Properties.Resources.MessageBoxCaptionFailure,
                MessageBoxButtons.OK,
                MessageBoxIcon.Hand,
                MessageBoxDefaultButton.Button1);
            Close();
        }

        /// <summary>
        /// Dictionary mapping each item type to the constructor of the
        /// edit control for that particular item type
        /// </summary>
        private static Dictionary<ItemType, ConstructorInfo> itemControlConstructors;

        /// <summary>
        /// Edit control created for the particular item we're editing
        /// </summary>
        private IEditItemControl editItemControl;

        private DataManager dataManager;
        private Item item;
        private ItemID folder;
    }
}