using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using PrivateDataCommon;
using PrivateDataCore;

namespace PrivateData
{
    public partial class ItemListForm : Form
    {
        private class ItemTreeNode : TreeNode
        {
            public ItemTreeNode(ItemID id, string name, ItemType type)
                : base(name)
            {
                this.id = id;
                this.type = type;
            }

            public ItemID ItemID
            {
                get
                {
                    return id;
                }
            }

            public ItemType ItemType
            {
                get
                {
                    return type;
                }
            }

            private ItemID id;
            private ItemType type;
        }

        static ItemListForm()
        {
            imageIndexes = new Dictionary<ItemType, int>();

            imageIndexes.Add(ItemType.BankAccount, 9);
            imageIndexes.Add(ItemType.CellPhone, 7);
            imageIndexes.Add(ItemType.CreditCard, 4);
            imageIndexes.Add(ItemType.Email, 6);
            imageIndexes.Add(ItemType.File, 5);
            imageIndexes.Add(ItemType.Folder, 1);
            imageIndexes.Add(ItemType.GenericPassword, 3);
            imageIndexes.Add(ItemType.Note, 8);
            imageIndexes.Add(ItemType.Picture, 11);
            imageIndexes.Add(ItemType.WebAccount, 2);
        }

        public ItemListForm(DataManager dataManager)
        {
            this.dataManager = dataManager;
            itemNodes = new Dictionary<ItemID, ItemTreeNode>();

            InitializeComponent();

            // Now that the TreeView exists, we can subscribe to the events
            dataManager.ItemStored
                += new EventHandler<DataManager.ItemEventArgs>(
                    dataManager_ItemStored);
            dataManager.ItemDeleted
                += new EventHandler<DataManager.ItemEventArgs>(
                    dataManager_ItemDeleted);
            dataManager.ItemTreeReloaded
                += new EventHandler<EventArgs>(dataManager_ItemTreeReloaded);
        }

        void dataManager_ItemTreeReloaded(object sender, EventArgs e)
        {
            // Refresh the itemTreeView on the UI thread. We will do this
            // synchronously because RefreshNode may need to query the data
            // manager for child data and it won't be good if the async operation
            // thread runs in the data manager (or storage) simultaneously.
            // Instead, we'll let the operation's thread wait while the
            // UI thread refreshes the TreeView.
            this.Invoke(
                new EventHandler<DataManager.ItemEventArgs>(RefreshNode),
                sender,
                new DataManager.ItemEventArgs(null));
        }

        private void ItemListForm_Load(object sender, EventArgs e)
        {
            FillItemTreeView();
        }

        private void FillItemTreeView()
        {
            root = new ItemTreeNode(
                null/*id*/,
                "Private Data Records", // TODO: From resource
                ItemType.Folder);
            root.BackColor = Color.Blue;
            root.ForeColor = Color.White;
            root.ImageIndex = 0;
            itemTreeView.Nodes.Add(root);

            AddChildren(root);
            root.Expand();
        }

        private ItemTreeNode CreateItemNode(ItemID item)
        {
            Metadata metadata = dataManager.GetMetadata(item);

            // Create and set up the item
            ItemTreeNode node = new ItemTreeNode(item, metadata.Name, metadata.Type);
            node.ImageIndex = imageIndexes[metadata.Type];
            itemNodes[item] = node;

            // If this is a folder, recursively build the tree
            if (metadata.Type == ItemType.Folder)
            {
                AddChildren(node);
            }

            // And return it
            return node;
        }

        private void AddChildren(ItemTreeNode node)
        {
            ItemID[] children
                = SortItems(dataManager.GetFolderItems(node.ItemID));
            foreach (ItemID child in children)
            {
                node.Nodes.Add(CreateItemNode(child));
            }
            node.Expand();
        }

        private class ItemMetadata : IComparable
        {
            public ItemMetadata(ItemID id, Metadata metadata)
            {
                this.id = id;
                this.metadata = metadata;
            }

            public int CompareTo(object obj)
            {
                if (obj is ItemMetadata)
                {
                    return CompareTo(obj as ItemMetadata);
                }

                throw new ArgumentException(); // TODO: Fix exceptions
            }

            public int CompareTo(ItemMetadata  other)
            {
                if (metadata.Type == ItemType.Folder
                    && other.metadata.Type != ItemType.Folder)
                {
                    return -1;
                }
                else if (metadata.Type != ItemType.Folder
                    && other.metadata.Type == ItemType.Folder)
                {
                    return 1;
                }
                else
                {
                    return metadata.Name.CompareTo(other.metadata.Name);
                }
            }

            public ItemID ID
            {
                get
                {
                    return id;
                }
            }

            private ItemID id;
            private Metadata metadata;
        }

        private ItemID[] SortItems(ItemID[] items)
        {
            ItemMetadata[] itemMetadata
                = new ItemMetadata[items.Length];
            int index = 0;
            foreach (ItemID item in items)
            {
                itemMetadata[index++]
                    = new ItemMetadata(item, dataManager.GetMetadata(item));
            }
            Array.Sort(itemMetadata);

            ItemID[] sortedIDs = new ItemID[items.Length];
            for (int i = 0; i < sortedIDs.Length; i++)
            {
                sortedIDs[i] = itemMetadata[i].ID;
            }

            return sortedIDs;
        }

        private void dataManager_ItemStored(object sender, DataManager.ItemEventArgs e)
        {
            // Refresh the item's node on the UI thread. We will do this
            // synchronously because RefreshNode may need to query the data
            // manager for child data and it won't be good if the async operation
            // thread runs in the data manager (or storage) simultaneously.
            // Instead, we'll let the operation's thread wait while the
            // UI thread refreshes the TreeView.
            this.Invoke(
                new EventHandler<DataManager.ItemEventArgs>(RefreshNode),
                sender,
                e);
        }

        private void dataManager_ItemDeleted(object sender, DataManager.ItemEventArgs e)
        {
            // Delete the item's node on the UI thread. Unlike RefreshNode,
            // DeleteNode does not need to query the data manager for more data,
            // but we'll execute it synchronously just to be consistent.
            this.Invoke(
                new EventHandler<DataManager.ItemEventArgs>(DeleteNode),
                sender,
                e);
        }

        private void RefreshNode(object sender, DataManager.ItemEventArgs e)
        {
            ItemID item = e.Item;

            ItemTreeNode node;
            bool itemInTree;
            if (item == null)
            {
                node = root;
                itemInTree = true;
            }
            else
            {
                itemInTree = itemNodes.TryGetValue(item, out node);
            }

            if (!itemInTree)
            {
                // Nothing to do here; this is a new item, so we'll eventually
                // get called to update its parent folder
            }
            else
            {
                // An already existing item (or the root) got changed
                SuspendLayout();

                Metadata metadata = (item == null ? null : dataManager.GetMetadata(item));

                // Update the node's name, if necessary
                if (metadata != null && node.Text != metadata.Name)
                {
                    node.Text = metadata.Name;
                }

                // And regenerate the children, if we're refreshing a folder
                if (item == null || metadata.Type == ItemType.Folder)
                {
                    node.Nodes.Clear();
                    AddChildren(node);
                }

                ResumeLayout();
            }
        }

        private void DeleteNode(object sender, DataManager.ItemEventArgs e)
        {
            if (itemNodes[e.Item].Parent.Nodes.Contains(itemNodes[e.Item]))
            {
                itemNodes[e.Item].Remove();
            }
            itemNodes.Remove(e.Item);
        }

        private void NewRecordMenuItem_Click(object sender, EventArgs e)
        {
            CreateNewItem(false);
        }

        private void NewFolderMenuItem_Click(object sender, EventArgs e)
        {
            CreateNewItem(true);
        }

        private void CreateNewItem(bool isFolder)
        {
            ItemID folder = SelectedFolder();
            CreateItemDialog createItemDialog
                = new CreateItemDialog(dataManager, folder, isFolder);
            createItemDialog.ShowDialog();
        }

        private ItemID SelectedFolder()
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;
            ItemID folder = node.ItemID;
            if (folder != null && node.ItemType != ItemType.Folder)
            {
                folder = ((ItemTreeNode)node.Parent).ItemID;
            }
            return folder;
        }

        private void itemTreeView_AfterSelect(object sender, TreeViewEventArgs e)
        {
            ItemTreeNode node = (ItemTreeNode)e.Node;
            itemTreeView.SelectedImageIndex = node.ImageIndex;

            if (node.ItemType == ItemType.Folder)
            {
                node.Expand();
            }

            viewContextMenuItem.Enabled = (node.ItemType != ItemType.Folder);
            viewMenuItem.Enabled = (node.ItemType != ItemType.Folder);
        }

        private void ViewItem(object sender, EventArgs e)
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;
            Item item;
            try
            {
                item = dataManager.GetItem(node.ItemID);
                EditItemDialog editItemDialog
                    = new EditItemDialog(
                        dataManager,
                        item,
                        ((ItemTreeNode)node.Parent).ItemID);
                editItemDialog.ShowDialog();
                editItemDialog.Dispose();
                editItemDialog = null;
            }
            catch (OutOfMemoryException exception)
            {
                // TODO: Show some appropriate message
                MessageBox.Show(exception.Message);
            }
        }

        private void RenameItem(object sender, EventArgs e)
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;

            if (node.ItemID == null)
            {
                // Internal error: why is Rename enabled for the root?
                throw new Exception(); // TODO: Fix exceptions
            }

            RenameItemDialog renameItemDialog = new RenameItemDialog(node.Text);
            if (renameItemDialog.ShowDialog() == DialogResult.OK)
            {
                if (renameItemDialog.NewName != node.Text)
                {
                    // TODO: Store only the modified matadata?
                    AsyncOperations.RenameItemOperation renameItem =
                        new AsyncOperations.RenameItemOperation(
                            dataManager,
                            node.ItemID,
                            renameItemDialog.NewName);

                    new AsyncOperationUI(
                        renameItem,
                        this,
                        new EventHandler(renameItem_Succeeded),
                        new EventHandler(renameItem_Cancelled),
                        new EventHandler(renameItem_Failed));

                    renameItem.BeginExecute();
                }
            }
        }

        private void renameItem_Succeeded(object sender, EventArgs e)
        {
            // Nothing to do here - the tree should have updated itself through
            // the events thrown from the data manager
        }

        private void renameItem_Cancelled(object sender, EventArgs e)
        {
        }

        private void renameItem_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Failed to rename item");
        }
        
        private void DeleteItem(object sender, EventArgs e)
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;

            if (node.ItemID == null)
            {
                // Internal error: why is Delete enabled for the root?
                throw new Exception(); // TODO: Fix exceptions
            }

            // TODO: Delete the item (ask for confirmation) and if the item is
            // a folder warn that every item inside it will be deleted as well
            string message;
            if (node.ItemType == ItemType.Folder)
            {
                message = "Are you sure you want to delete this folder and all of its contents?";
            }
            else
            {
                message = "Are you sure you want to delete this item?";
            }

            if (MessageBox.Show(
                    message, "Delete item?", // TODO: From resource
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation,
                    MessageBoxDefaultButton.Button2)
                == DialogResult.Yes)
            {
                //Delete the item
                AsyncOperations.DeleteItemOperation deleteItem =
                    new AsyncOperations.DeleteItemOperation(
                        dataManager,
                        node.ItemID);

                new AsyncOperationUI(
                    deleteItem,
                    this,
                    new EventHandler(deleteItem_Succeeded),
                    new EventHandler(deleteItem_Cancelled),
                    new EventHandler(deleteItem_Failed));

                deleteItem.BeginExecute();
            }
        }

        private void deleteItem_Succeeded(object sender, EventArgs e)
        {
            // Nothing to do here - the tree should have updated itself through
            // the events thrown from the data manager
        }

        private void deleteItem_Cancelled(object sender, EventArgs e)
        {
            // This operation should not have been cancellable
            throw new Exception(); // TODO: Fix exceptions
        }

        private void deleteItem_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Failed to delete item");
        }

        private void CutItem(object sender, EventArgs e)
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;

            if (node.ItemID == null)
            {
                // Internal error: why was Cut allowed on the root?
                throw new Exception(); // TODO: Fix exceptions
            }

            cutItem = node.ItemID;
        }

        private void PasteItem(object sender, EventArgs e)
        {
            if (cutItem == null)
            {
                // Internal error: why is Paste enabled when nothing is cut?
                throw new Exception(); // TODO: Fix exceptions
            }

            ItemID folder = SelectedFolder();

            AsyncOperations.MoveItemToFolderOperation moveItem
                = new AsyncOperations.MoveItemToFolderOperation(
                    dataManager,
                    cutItem,
                    folder);

            // Do this now, so nobody can call Paste while it's executing (just in case)
            cutItem = null;

            new AsyncOperationUI(
                moveItem,
                this,
                new EventHandler(MoveItem_Succeeded),
                new EventHandler(MoveItem_Cancelled),
                new EventHandler(MoveItem_Failed));

            moveItem.BeginExecute();
        }

        private void MoveItem_Succeeded(object sender, EventArgs e)
        {
            // Nothing to do here - the tree should have updated itself through
            // the events thrown from the data manager
        }

        private void MoveItem_Cancelled(object sender, EventArgs e)
        {
            // This operation should not be cancellable
            throw new Exception(); // TODO: Fix exceptions
        }

        private void MoveItem_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Failed to move item");
        }

        private bool CutEnabled()
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;

            // Only the root can't be cut (and an already cut item can't be re-cut)
            return (node.ItemID != null && node.ItemID != cutItem);
        }

        private bool PasteEnabled()
        {
            // If something has been cut - it can be pasted (unless on top of itself)
            return (cutItem != null && SelectedFolder() != cutItem);
        }

        private bool RenameEnabled()
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;

            // Only the root can't be renamed
            return (node.ItemID != null);
        }

        private bool DeleteEnabled()
        {
            ItemTreeNode node = (ItemTreeNode)itemTreeView.SelectedNode;

            // Only the root can't be deleted
            return (node.ItemID != null);
        }

        private void contextMenu_Popup(object sender, EventArgs e)
        {
            cutContextMenuItem.Enabled = CutEnabled();
            pasteContextMenuItem.Enabled = PasteEnabled();
            renameContextMenuItem.Enabled = RenameEnabled();
            deleteContextMenuItem.Enabled = DeleteEnabled();
        }

        private void editMenuItem_Popup(object sender, EventArgs e)
        {
            editCutMenuItem.Enabled = CutEnabled();
            editPasteMenuItem.Enabled = PasteEnabled();
            editRenameMenuItem.Enabled = RenameEnabled();
            editDeleteMenuItem.Enabled = DeleteEnabled();
        }

        private void Defragment_Succeeded(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("File compacting finished.");
        }

        private void Defragment_Cancelled(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("File compacting cancelled.");
        }

        private void Defragment_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("File compacting failed.");
        }

        private void compactFileMenuItem_Click(object sender, EventArgs e)
        {
            AsyncOperations.CompactFileOperation defragment
                = new AsyncOperations.CompactFileOperation(dataManager);

            new AsyncOperationUI(
                defragment,
                this,
                new EventHandler(Defragment_Succeeded),
                new EventHandler(Defragment_Cancelled),
                new EventHandler(Defragment_Failed));

            defragment.BeginExecute();
        }

        private void Backup_Succeeded(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Your data was successfully backed up.");
        }

        private void Backup_Cancelled(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Data backup was cancelled.");
        }

        private void Backup_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Data backup failed.");
        }

        private void backupMenuItem_Click(object sender, EventArgs e)
        {
            // Show the saveFileDialog
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "Private Data Archive (*.pda)|*.pda";
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                // Ask for a password with NewPasswordDialog
                NewPasswordDialog newPasswordDialog
                    = new NewPasswordDialog(false);
                if (newPasswordDialog.ShowDialog() == DialogResult.OK)
                {
                    // Create the archive
                    AsyncOperations.ArchiveOperation archive
                        = new AsyncOperations.ArchiveOperation(
                            dataManager,
                            saveFileDialog.FileName,
                            newPasswordDialog.NewPassword,
                            2500); // TODO: set proper iteration count

                    new AsyncOperationUI(
                        archive,
                        this,
                        new EventHandler(Backup_Succeeded),
                        new EventHandler(Backup_Cancelled),
                        new EventHandler(Backup_Failed));

                    archive.BeginExecute();
                }
            }
        }

        private void Restore_Succeeded(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Your data was successfully restored.");
        }

        private void Restore_Cancelled(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Data restore was cancelled.");
        }

        private void Restore_Failed(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Data restore failed.");
        }

        private void restoreMenuItem_Click(object sender, EventArgs e)
        {
            // TODO: Implement
            // Show the openFileDialog
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "Private Data Archive (*.pda)|*.pda";
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                // Ask for a password
                BackupPasswordDialog backupPasswordDialog
                    = new BackupPasswordDialog();
                backupPasswordDialog.FileName
                    = System.IO.Path.GetFileName(openFileDialog.FileName);
                if (backupPasswordDialog.ShowDialog() == DialogResult.OK)
                {
                    // Atempt to restore from the file
                    AsyncOperations.RestoreOperation restore
                        = new AsyncOperations.RestoreOperation(
                            dataManager,
                            openFileDialog.FileName,
                            backupPasswordDialog.Password);

                    new AsyncOperationUI(
                        restore,
                        this,
                        new EventHandler(Restore_Succeeded),
                        new EventHandler(Restore_Cancelled),
                        new EventHandler(Restore_Failed));

                    restore.BeginExecute();
                }
            }
        }

        private void changePasswordMenuItem_Click(object sender, EventArgs e)
        {
            NewPasswordDialog newPasswordDialog
                = new NewPasswordDialog(true/*changePassword*/);
            if (newPasswordDialog.ShowDialog() == DialogResult.OK)
            {
                AsyncOperations.ChangePasswordOperation changePassword
                    = new PrivateData.AsyncOperations.ChangePasswordOperation(
                        dataManager,
                        newPasswordDialog.OldPassword,
                        newPasswordDialog.NewPassword,
                        2500); // TODO: set proper iteration count

                new AsyncOperationUI(
                    changePassword,
                    this,
                    new EventHandler(ChangePassword_Succeeded),
                    new EventHandler(ChangePassword_Cancelled),
                    new EventHandler(ChangePassword_Failed));

                changePassword.BeginExecute();
            }
        }

        private void ChangePassword_Succeeded(object sender, EventArgs e)
        {
            // TODO: From resource
            MessageBox.Show("Your password has been changed successfully.");
        }

        private void ChangePassword_Cancelled(object sender, EventArgs e)
        {
            // This operation should not be cancellable
            throw new Exception(); // TODO: Fix exceptions
        }

        private void ChangePassword_Failed(object sender, EventArgs e)
        {
            AsyncOperations.ChangePasswordOperation changePassword
                = sender as AsyncOperations.ChangePasswordOperation;

            if (changePassword.IncorrectPassword)
            {
                // TODO: Handle wrong password (set label?)
                MessageBox.Show("The password you entered was incorrect. If you want, you can try to change the password again.");
            }
            else
            {
                // TODO: From resource
                MessageBox.Show("Your password was not changed.");
            }
        }

        private static Dictionary<ItemType, int> imageIndexes;

        private DataManager dataManager;

        private ItemTreeNode root;
        private Dictionary<ItemID, ItemTreeNode> itemNodes;

        private ItemID cutItem;
    }
}