using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using PrivateDataCommon;
using PrivateDataStorage;

namespace TestPrivateData.Tests
{
    class TestDataStorage : Test
    {
        public override string Name
        {
            get
            {
                return "Data Storage";
            }
        }

        public override bool ExecuteTest(Test.SetTestStatusDelegate setStatus)
        {
            const int customDataItemCount = 10;
            const int itemCount = 500;
            const string dateFormat = "{0:yyyy}-{0:MM}-{0:dd} {0:HH}{0:mm}{0:ss}";

            DateTime testDateTime = DateTime.Now;

            // Create a new storage
            string storagePath
                = Path.Combine(
                        ".", // TODO: Write in a proper location
                        String.Format("Storage " + dateFormat + ".PDS", testDateTime));

            setStatus("Creating new storage...");
            DataStorage storage = DataStorage.CreateStorage(storagePath);

            try
            {
#if MEASURING_TIME
                // Get a list of storage cards; we'll use the first
                string[] storageCards = GetStorageCardPaths();
                if (storageCards.Length < 1)
                {
                    throw new Exception(); // TODO: Fix exceptions
                }

                FileStream perfFile
                    = new FileStream(
                        Path.Combine(
                            storageCards[0],
                            String.Format(
                                "Perf " + dateFormat + ".txt",
                                testDateTime)),
                        FileMode.CreateNew,
                        FileAccess.Write,
                        FileShare.Read);
                StreamWriter perfWriter = new StreamWriter(perfFile);
#endif

                // Create some random custom data items to store
                TestCustomDataItem[] testCustomDataItemsByType;
                {
                    setStatus("Creating random test custom data items...");
                    TestCustomDataItem[] testCustomDataItems
                        = CreateRandomTestCustomDataItems(customDataItemCount);
                    List<TestCustomDataItem> testCustomDataItemsByTypeList =
                        new List<TestCustomDataItem>();

                    // Store the custom data items
                    for (int i = 0; i < testCustomDataItems.Length; i++)
                    {
                        // Store the custom data item in the storage
                        StoreTestCustomDataItem(storage, testCustomDataItems[i]);

                        // If the list of items by type is not large enough,
                        // expand it
                        if (testCustomDataItemsByTypeList.Count <= testCustomDataItems[i].type)
                        {
                            for (int t = testCustomDataItemsByTypeList.Count;
                                 t <= testCustomDataItems[i].type;
                                 t++)
                            {
                                testCustomDataItemsByTypeList.Add(null);
                            }
                        }

                        // File the item according to its type so we can verify it
                        // later (only the last item of a particular type is kept
                        // around, just like the storage does)
                        testCustomDataItemsByTypeList[testCustomDataItems[i].type]
                            = testCustomDataItems[i];
                    }

                    testCustomDataItemsByType
                        = testCustomDataItemsByTypeList.ToArray();
                }

                // Create some random items to store
                setStatus("Creating random test items...");
                TestItem[] testItems = CreateRandomTestItems(itemCount);

                // Store the items
                Timer timer = new Timer();
                for (int i = 0; i < testItems.Length; i++)
                {
                    setStatus(
                        String.Format("Storing item {0} of {1} ({2})...",
                                      i + 1,
                                      testItems.Length,
                                      testItems[i].isExternal ? "external" : "internal"));

#if MEASURING_TIME
                    timer.Start();
#endif

                    StoreTestItem(
                        storage,
                        testItems[i],
                        new SlowOperationController());

#if MEASURING_TIME
                    long millisecondsToStore = timer.Stop();
                    perfWriter.WriteLine(millisecondsToStore);
#endif
                }

#if MEASURING_TIME
                perfWriter.Close(); // Also closes the underlying stream
#endif

                // Randomly delete/modify/create some items
                Random random = new Random();
                int operationCount = random.Next(testItems.Length * 2);
                bool[] deleted = new bool[testItems.Length];
                List<TestItem> createdItems = new List<TestItem>();

                for (int i = 0; i < operationCount; i++)
                {
                    int operation = random.Next(3);
                    switch (operation)
                    {
                        case 0:
                            // Delete
                            {
                                int index = RandomNotDeletedItem(random, deleted);

                                if (index >= 0)
                                {
                                    setStatus(
                                        String.Format("Deleting item {0}...",
                                                      index + 1));
                                    storage.DeleteItem(testItems[index].id);
                                    deleted[index] = true;
                                }
                            }
                            break;

                        case 1:
                            // Modify
                            {
                                int index = RandomNotDeletedItem(random, deleted);

                                if (index >= 0)
                                {
                                    setStatus(
                                        String.Format("Modifying item {0}...",
                                                      index + 1));

                                    // Create a new item, but save the ID
                                    {
                                        ItemID id = testItems[index].id;
                                        testItems[index] = CreateRandomTestItem(random);
                                        testItems[index].id = id;
                                    }

                                    StoreTestItem(
                                        storage,
                                        testItems[index],
                                        new SlowOperationController());
                                }
                            }
                            break;

                        case 2:
                            // Create
                            {
                                setStatus("Creating new item...");
                                TestItem newItem = CreateRandomTestItem(random);
                                createdItems.Add(newItem);
                                StoreTestItem(
                                    storage,
                                    newItem,
                                    new SlowOperationController());
                            }
                            break;
                    }
                }

                // Read back the data and verify it's still correct
                if (!VerifyData(
                        setStatus,
                        storage,
                        testCustomDataItemsByType,
                        testItems,
                        deleted,
                        createdItems))
                {
                    return false;
                }

                // Close and reopen the storage and verify that the data
                // still reads back correctly
                setStatus("Closing file...");
                storage.Close();
                storage = null; // So that we don't call Close on an invalid
                // object in the "finally" block if something
                // goes wrong
                setStatus("Reopening file...");
                storage = new DataStorage(storagePath);
                if (!VerifyData(
                        setStatus,
                        storage,
                        testCustomDataItemsByType,
                        testItems,
                        deleted,
                        createdItems))
                {
                    return false;
                }

                setStatus("Defragmenting...");
                storage.Defragment(new SlowOperationController());
                if (!VerifyData(
                        setStatus,
                        storage,
                        testCustomDataItemsByType,
                        testItems,
                        deleted,
                        createdItems))
                {
                    return false;
                }

                // Now that we have defragmented, close and reopen the storage
                // and verify that the data still reads back correctly
                setStatus("Closing file...");
                storage.Close();
                storage = null; // So that we don't call Close on an invalid
                // object in the "finally" block if something goes wrong
                setStatus("Reopening file...");
                storage = new DataStorage(storagePath);
                if (!VerifyData(
                        setStatus,
                        storage,
                        testCustomDataItemsByType,
                        testItems,
                        deleted,
                        createdItems))
                {
                    return false;
                }
            }
            finally
            {
                if (storage != null)
                {
                    storage.Close();
                    storage = null;
                }
            }

            return true;
        }

        /// <summary>
        /// Stores a test custom data item into the storage
        /// </summary>
        /// <param name="storage">Data storage</param>
        /// <param name="item">Contents to store</param>
        private static void StoreTestCustomDataItem(
                                DataStorage storage,
                                TestCustomDataItem item)
        {
            storage.StoreCustomData(
                item.type,
                item.contentIV != null
                    ? new EncryptedContent(
                        new MemoryStream(item.content), item.contentIV)
                    : new Content(
                        new MemoryStream(item.content)));
        }

        /// <summary>
        /// Stores a test item into the storage
        /// </summary>
        /// <param name="storage">Data storage</param>
        /// <param name="item">Contents to store</param>
        /// <param name="controller">Slow operation controller</param>
        /// <returns>ID of the item</returns>
        private static void StoreTestItem(
            DataStorage storage,
            TestItem item,
            SlowOperationController controller)
        {
            storage.StoreItem(
                item.id,
                new EncryptedContent(
                    new MemoryStream(item.metadata),
                    item.metadataIV),
                new EncryptedContent(
                    new MemoryStream(item.content),
                    item.contentIV),
                item.isExternal,
                controller);
        }

        private static bool VerifyData(
            SetTestStatusDelegate setStatus,
            DataStorage storage,
            TestCustomDataItem[] testCustomDataItemsByType,
            TestItem[] testItems,
            bool[] deleted,
            List<TestItem> createdItems)
        {
            const string verifyingCustomData = "Verifying custom data of type {0}...";
            const string verifyingItem = "Verifying item {0} of {1}...";
            const string verifyingItemDeleted = "Verifying item {0} of {1} (deleted)...";

            // Verify the custom data
            for (int type = 0; type < testCustomDataItemsByType.Length; type++)
            {
                TestCustomDataItem customDataItem = testCustomDataItemsByType[type];

                // If we have a custom data item for this type, verify it
                if (customDataItem != null)
                {
                    setStatus(String.Format(verifyingCustomData, type));

                    if (customDataItem.type != type)
                    {
                        // Internal error - why is this custom data item filed
                        // incorrectly?
                        throw new Exception();  // TODO: Fix exceptions
                    }

                    // Verify that the contents match
                    Content customDataContent = storage.GetCustomData(type);
                    try
                    {
                        if (customDataItem.contentIV != null)
                        {
                            // We have encrypted content
                            EncryptedContent encryptedContent
                                = customDataContent as EncryptedContent;

                            if (!VerifyEncryptedContent(
                                    encryptedContent,
                                    customDataItem.content,
                                    customDataItem.contentIV))
                            {
                                return false;
                            }
                        }
                        else
                        {
                            // We have non-encrypted content

                            // First, check that the storage also thinks the
                            // custom data item is not encrypted
                            if (customDataContent.GetType() == typeof(EncryptedContent))
                            {
                                return false;
                            }

                            // Then, check the contents
                            if (!VerifyContent(
                                    customDataContent,
                                    customDataItem.content))
                            {
                                return false;
                            }
                        }
                    }
                    finally
                    {
                        customDataContent.Dispose();
                    }
                }
            }

            int current = 0;
            int count = testItems.Length + createdItems.Count;

            for (int i = 0; i < testItems.Length; i++)
            {
                bool verifyOK;

                if (!deleted[i])
                {
                    setStatus(String.Format(verifyingItem, ++current, count));
                    verifyOK = VerifyItem(storage, testItems[i]);
                }
                else
                {
                    setStatus(String.Format(verifyingItemDeleted, ++current, count));
                    verifyOK = VerifyDeleted(storage, testItems[i]);
                }

                if (!verifyOK)
                {
                    return false;
                }
            }

            for (int i = 0; i < createdItems.Count; i++)
            {
                setStatus(String.Format(verifyingItem, ++current, count));
                if (!VerifyItem(storage, createdItems[i]))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Verifies that the storage returns the correct data for a test item
        /// </summary>
        /// <param name="storage">Data storage</param>
        /// <param name="item">Expected contents</param>
        /// <returns>
        /// True if the storage's data matches the expected contents; otherwise false
        /// </returns>
        private static bool VerifyItem(DataStorage storage, TestItem item)
        {
            {
                EncryptedContent metadata = storage.GetMetadata(item.id);
                if (!VerifyEncryptedContent(metadata, item.metadata, item.metadataIV))
                {
                    return false;
                }
                metadata.Dispose();
            }

            {
                EncryptedContent content = storage.GetContent(item.id);
                if (!VerifyEncryptedContent(content, item.content, item.contentIV))
                {
                    return false;
                }
                content.Dispose();
            }

            return true;
        }

        private static bool VerifyContent(Content content, byte[] data)
        {
            Stream contentStream = content.DetachData();
            byte[] contentBytes = Utils.ReadStream(contentStream);
            contentStream.Close();

            return Utils.ByteArraysMatch(contentBytes, data);
        }

        private static bool VerifyEncryptedContent(EncryptedContent encryptedContent,
                                                   byte[] data,
                                                   IV iv)
        {
            return VerifyContent(encryptedContent, data)
                && Utils.ByteArraysMatch(encryptedContent.IV.Bytes, iv.Bytes);
        }

        /// <summary>
        /// Verifies that a test item has indeed been deleted
        /// </summary>
        /// <param name="storage">Data storage</param>
        /// <param name="item">Item to verify</param>
        /// <returns>
        /// True if the item has indeed been deleted and is inaccessible;
        /// otherwise false
        /// </returns>
        private static bool VerifyDeleted(DataStorage storage, TestItem item)
        {
            bool deleted = false;

            try
            {
                EncryptedContent readContent = storage.GetContent(item.id);
                readContent.Dispose();
            }
            catch // TODO: Fix exceptions
            {
                deleted = true;
            }

            return deleted;
        }

        /// <summary>
        /// Returns the random index of a not-deleted item
        /// </summary>
        /// <param name="random">Random to use</param>
        /// <param name="deleted">Array signifying which items are deleted</param>
        /// <returns>
        /// Index of a random not-deleted item, or -1 if all items have been deleted
        /// </returns>
        private static int RandomNotDeletedItem(Random random, bool[] deleted)
        {
            int index = random.Next(deleted.Length);

            // If the randomly chosen item has been deleted already, look forward
            // for the first not-yet-deleted item
            if (deleted[index])
            {
                for (int i = index; i < deleted.Length; i++)
                {
                    if (!deleted[i])
                    {
                        index = i;
                        break;
                    }
                }
            }

            // If the forward search was unsuccessful, continue searching from
            // the beginning
            if (deleted[index])
            {
                for (int i = 0; i < index; i++)
                {
                    if (!deleted[i])
                    {
                        index = i;
                        break;
                    }
                }
            }

            return !deleted[index] ? index : -1;
        }

        /// <summary>
        /// Creates random test custom data items
        /// </summary>
        /// <param name="count">Number of custom data items to create</param>
        /// <returns>Random test custom data items</returns>
        private static TestCustomDataItem[] CreateRandomTestCustomDataItems(int count)
        {
            Random random = new Random();
            TestCustomDataItem[] testCustomDataItems
                = new TestCustomDataItem[count];

            // Generate the test custom data items
            for (int i = 0; i < count; i++)
            {
                testCustomDataItems[i]
                    = CreateRandomTestCustomDataItem(random, count - 1);
            }

            return testCustomDataItems;
        }

        /// <summary>
        /// Creates random test items
        /// </summary>
        /// <param name="count">Number of items to create</param>
        /// <returns>Random test items</returns>
        private static TestItem[] CreateRandomTestItems(int count)
        {
            Random random = new Random();
            TestItem[] testItems = new TestItem[count];

            // Generate the test items
            for (int i = 0; i < count; i++)
            {
                testItems[i] = CreateRandomTestItem(random);
            }

            return testItems;
        }

        /// <summary>
        /// Creates a single random test item
        /// </summary>
        /// <param name="random">Random to use</param>
        /// <param name="maxType">Largest possible value for the custom data type</param>
        /// <returns>Random test item</returns>
        private static TestCustomDataItem CreateRandomTestCustomDataItem(Random random, int maxType)
        {
            const int maxIVLength = 20;
            const int maxContentLength = 500;
            const int encryptedChance = 2; // 1/encryptedChance items are encrypted

            TestCustomDataItem item = new TestCustomDataItem();

            // Create the content
            {
                item.type = random.Next(maxType + 1);
                item.content = CreateRandomTestData(random, maxContentLength);

                // Encrypted custom data item?
                if (random.Next(encryptedChance) == 0)
                {
                    item.contentIV = new IV(CreateRandomTestData(random, maxIVLength));
                }
            }

            return item;
        }

        /// <summary>
        /// Creates a single random test item
        /// </summary>
        /// <param name="random">Random to use</param>
        /// <returns>Random test item</returns>
        private static TestItem CreateRandomTestItem(Random random)
        {
            const int maxIVLength = 20;
            const int maxMetadataLength = 150;
            const int maxContentLengthExternal = 10240; // 10K max
            const int maxContentLengthInternal = 500;
            const int internalToExternalRatio = 2; // internal:external = 2:1
            // 1/20th chance that we store a big item internally or a small
            // item externally
            const int incorrectFilingChance = 20;

            TestItem item = new TestItem();

            // Generate a new ID
            item.id = new ItemID();

            // Create the metadata
            {
                item.metadata = CreateRandomTestData(random, maxMetadataLength);
                item.metadataIV = new IV(CreateRandomTestData(random, maxIVLength));
            }

            // External?
            item.isExternal
                = (random.Next(internalToExternalRatio + 1) == 0
                    ? true
                    : false);

            // Create the content
            {
                item.content
                    = CreateRandomTestData(random,
                                              item.isExternal
                                                  ? maxContentLengthExternal
                                                  : maxContentLengthInternal);
                item.contentIV = new IV(CreateRandomTestData(random, maxIVLength));
            }

            // 1/incorrectFilingChance chance that we store a big item
            // internally or a small item externally
            if (random.Next(incorrectFilingChance) == 0)
            {
                item.isExternal = !item.isExternal;
            }

            return item;
        }

        /// <summary>
        /// Creates a random byte array
        /// </summary>
        /// <param name="random">Random to use (for performance reasons)</param>
        /// <param name="maxLength">Maximum length of the stream</param>
        /// Verifier test stream (same contents as the test stream)
        /// </param>
        /// <returns>Byte array generated</returns>
        private static byte[] CreateRandomTestData(Random random,
                                                   int maxLength)
        {
            byte[] bytes = new byte[random.Next(maxLength) + 1];
            random.NextBytes(bytes);

            return bytes;
        }

        private class TestCustomDataItem
        {
            public int type;
            public byte[] content;
            public IV contentIV; // May be null if this is not an encrypted item
        }

        private class TestItem
        {
            public ItemID id;
            public byte[] metadata;
            public IV metadataIV;
            public byte[] content;
            public IV contentIV;
            public bool isExternal;
        }

        /// <summary>
        /// Compares two EncryptedContent objects. Note that it detaches the
        /// data, essentially destroying the objects to be compared.
        /// </summary>
        /// <param name="content1">First EncryptedContent to compare</param>
        /// <param name="content2">Second EncryptedContent to compare</param>
        /// <returns>True if the objects match. Otherwise false.</returns>
        private static bool EncryptedContentsMatch(EncryptedContent content1,
                                                   EncryptedContent content2)
        {
            if (!Utils.ByteArraysMatch(content1.IV.Bytes, content2.IV.Bytes))
            {
                return false;
            }

            Stream content1Data = content1.DetachData();
            byte[] content1Bytes = Utils.ReadStream(content1Data);
            content1Data.Close();

            Stream content2Data = content2.DetachData();
            byte[] content2Bytes = Utils.ReadStream(content2Data);
            content2Data.Close();

            return Utils.ByteArraysMatch(content1Bytes, content2Bytes);
        }

        /// <summary>
        /// Returns a list of paths to the mobile device's storage cards
        /// </summary>
        /// <returns>List of full paths to the mobile device's storage cards</returns>
        private static string[] GetStorageCardPaths()
        {
            const FileAttributes StorageCardAttributes
                = (FileAttributes.Directory | FileAttributes.Temporary);

            List<string> storageCards = new List<string>();
            DirectoryInfo rootDir = new DirectoryInfo(@"\");

            foreach (DirectoryInfo directoryInfo in rootDir.GetDirectories())
            {
                // If a temporary directory
                if ((directoryInfo.Attributes & StorageCardAttributes)
                    == StorageCardAttributes)
                {
                    storageCards.Add(directoryInfo.FullName);
                }
            }

            return storageCards.ToArray();
        }
    }
}
